import numpy as np
import scipy.sparse as sparse
import scipy.sparse.linalg as splg
from numpy import exp, fft, pi, sqrt
from scipy.constants import c as c0
from scipy.constants import e as e0
from scipy.constants import electron_mass as m0
from scipy.constants import epsilon_0 as eps0
from scipy.constants import h as h
from scipy.constants import hbar as hbar
from scipy.linalg import null_space

try:
    from ErwinJr2.OneDQuantum import c_schrodinger as onedq
except OSError:
    onedq = None
    print('C library is not compiled. Features are limited.')
import copy
from typing import List, Tuple, Union

from ErwinJr2 import Material

EUNIT = 1e-5    # E field unit from kV/cm to V/Angstrom
BASISPAD = 100  #: padding barrier for basis solver, unit Angstrom
INV_INF = 1e-20  #: for infinite small decay rate (ns-1)
# used for typing as either an array or a float number
ScalerOrArray = Union[float, np.ndarray]

QCMaterial = {
    "InP":  ["InGaAs", "AlInAs"],
    "GaAs": ["AlGaAs"],
    "GaSb": ["InAsSb", "AlGaSb"]
}  #: supported substrate/material set


class StateRecognizeError(Exception):
    """Raised when QClayers cannot recognize a period of states.
    This typically means the number of periods is too small, or for single
    period structure the barrier at the end of the structure is not long
    enough to bound a state.

    Attributes:
        expression -- input expression in which the error occurred."""
    def __init__(self, expression=''):
        self.expression = expression


class SchrodingerLayer(object):
    """Class for layer structure for Schrodinger solver.

    This is used as the base class of :class:`.QCLayers` for separation of
    material property and the solver.

Parameters
----------
xres :
    Position resolution, in Armstrong
Eres :
    Energy resolution, in meV.
    This number being too large may results in miss of some states while
    this being too small will make a long computation time.
    The parameter does not mean the accuracy of the eigen energy. It's
    required for algorithm reasons because of lack of a universal global
    root finding.
statePerRepeat :
    Number of states per repeat, used for calculating matrixEigenCount

layerWidths :
    Width of each layer, in angstrom. len = No. of layers
layerVc :
    The conduction band offset of each layer, in unit eV, len = No. of layers
layerMc :
    The conduction band effective mass of each layer, in unit m0,
    the free space electron mass, len = No. of layers
layerARs :
    Binaries indicating if the layer is active(True) or not(False),
    only affects basis solver, len = No. of layers

ifrDelta :
    The standard deviation of the interface roughness for the interface at
    layer n and layer n+1, in unit angstrom, len = No. of layers.
    Default zero
ifrLambda :
    The correlation length of the interface roughness for the interface at
    layer n and layer n+1, in unit angstrom, len = No. of layers.
    Default zero
avghwLO :
    The average LO phonon energy in unit eV.
    This value is used for LO phonon scattering calculation.
epsrho :
    The effective relative permittivity for LO phonon.
    1/epsrho = 1/(epsilon for high frequency) - 1/(epsilon for static)
    The value is used for LO phonon scattering calculation.

EField :
    External (static) electrical field, in kV/cm = 1e5 V/m
repeats :
    Number of repeat times for the given structure

crystalType :
    Default being "simple", meaning a simple parabolic effective mass is
    used for the calculation. For setting other than "simple",
    `populate_material` should be implemented.

basisAROnly :
    For basis solver if only the Active Region (AR) should be solved.
basisARInjector :
    For basis solver if there should be separator between AR->Injector
basisInjectorAR :
    For basis solver if there should be separator between Injector->AR

solver :
    The solver used for the eigen problem: 'ODE' or 'matrix'.
    By default 'ODE' if C library exists, 'matrix' is a full back.
includeIFR :
    Weather to include IFR scattering for performance estimation.
matrixEigenCount :
    The number of eigen pairs to calculate in the 'matrix' solver.
    It would be very expensive to calculate all of them.

status :
    - 'unsolved' meaning the structure is not solved yet.
    - 'basis' meaning the eigen problem is solved for basis
    - 'solved' meaning the eigen problem is solved.
    - 'solved-full' meaning the population distribution is known.
    """
    xres: float
    Eres: float
    statePerRepeat: int
    layerWidths: List[float]
    _layerVc: List[float]
    _layerMc: List[float]
    layerARs: List[bool]
    ifrDelta: List[float]
    ifrLambda: List[float]
    avghwLO: float
    epsrho: float
    EField: float
    repeats: int
    crystalType: str
    basisARInjector: bool
    basisInjectorAR: bool
    basisAROnly: bool
    matrixEigenCount: int

    solver: str
    includeIFR: bool

    status: str

    def __init__(self, xres: float = 0.5, Eres: float = 0.5,
                 statePerRepeat: int = 20,
                 layerWidths: List[float] = [10.0],
                 layerARs: List[bool] = None,
                 layerVc: List[float] = None,
                 layerMc: List[float] = None,
                 ifrDelta: List[float] = None,
                 ifrLambda: List[float] = None,
                 avghwLO: float = 35E-3,
                 epsrho: float = 69.0,
                 EField: float = 0.0, repeats: int = 1):
        self.xres = xres
        self.Eres = Eres
        self.statePerRepeat = statePerRepeat
        assert isinstance(layerWidths, list)
        self.layerWidths = layerWidths
        N = len(layerWidths)
        self._layerVc = layerVc if layerVc is not None else [0.0]*N
        self._layerMc = layerMc if layerMc is not None else [1.0]*N
        self.layerARs = layerARs if layerARs is not None else [True]*N
        self.ifrDelta = ifrDelta if ifrDelta is not None else [0.0]*N
        self.ifrLambda = ifrLambda if ifrLambda is not None else [0.0]*N
        self.EField = EField
        self.repeats = repeats
        self.avghwLO = avghwLO
        self.epsrho = epsrho

        self.crystalType = 'simple'
        self.solver = 'ODE'
        self.includeIFR = False
        self.matrixEigenCount = statePerRepeat * repeats

        self.basisAROnly = False
        self.basisARInjector = True
        self.basisInjectorAR = True

        self.status = 'unsolved'

    def layerVc(self, n: int) -> float:
        """The conduction band offset at n-th layer in eV"""
        return self._layerVc[n]

    def layerMc(self, n: int) -> float:
        """The conduction band effective mass at n-th layer, in m0"""
        return self._layerMc[n]

    def rotate_layer(self):
        for layerList in (self.layerWidths, self.layerARs,
                          self._layerVc, self._layerMc,
                          self.ifrDelta, self.ifrLambda):
            layerList.insert(0, layerList.pop())
        self.status = 'unsolved'

    def del_layer(self, n: int):
        for layerList in (self.layerWidths, self.layerARs,
                          self._layerVc, self._layerMc,
                          self.ifrDelta, self.ifrLambda):
            layerList.pop(n)
        self.status = 'unsolved'

    def add_layer(self, n: int, width: float,
                  Vc: float = 0.0, mc: float = 1.0, AR: bool = True,
                  ifrDelta: float = 0.0, ifrLambda: float = 0.0):
        self.layerWidths.insert(n, width)
        self.layerARs.insert(n, AR)
        self._layerVc.insert(n, Vc)
        self._layerMc.insert(n, mc)
        self.ifrDelta.insert(n, ifrDelta)
        self.ifrLambda.insert(n, ifrLambda)
        self.status = 'unsolved'

    def invert_layer(self):
        self.layerWidths = self.layerWidths[::-1]
        self._layerVc = self._layerVc[::-1]
        self._layerMc = self._layerMc[::-1]
        self.layerARs = self.layerARs[::-1]
        self.ifrDelta = self.ifrDelta[:-1:-1] + [self.ifrDelta[-1]]
        self.ifrLambda = self.ifrLambda[:-1:-1] + [self.ifrLambda[-1]]
        self.status = 'unsolved'

    def populate_x(self):
        """Calculate the properties in terms of position

        Yield
        -----
        xPoints : np.ndarray of float
            position grid
        xLayerNums : np.ndarray of int
            at xPoints[i] it's xLayerNums[i]-th layer
        xVc : np.ndarray of float
            The band offset energy at each point
        xMc : np.ndarray of float
            The effective mass at different points
        """
        layerCumSum = [0] + np.cumsum(self.layerWidths).tolist()
        self.periodL = layerCumSum[-1]
        self.xPoints = np.arange(0, self.periodL*self.repeats, self.xres)
        N = self.xPoints.size
        if N == 0:
            self.xPoints = np.array([0])
            N = 1
        self.xLayerNums = np.empty(N, dtype=int)
        self.xVc = np.empty(N)
        self.xMc = np.empty(N)

        for n in range(len(self.layerWidths)):
            indices = np.logical_or.reduce([
                (self.xPoints >= layerCumSum[n] + k * self.periodL)
                & (self.xPoints < layerCumSum[n+1] + k * self.periodL)
                for k in range(self.repeats)])
            self.xLayerNums[indices] = n
            self.xVc[indices] = self.layerVc(n)
            self.xMc[indices] = self.layerMc(n)
        if self.crystalType != 'simple':
            self.populate_material()

        self.offset = np.max(self.xVc) - np.min(self.xVc)
        self.xVField = self.xPoints * self.EField * EUNIT
        self.xVc -= self.xVField

        self.Es = np.arange(np.min(self.xVc), np.max(self.xVc), self.Eres/1E3)
        self.matrixEigenCount = self.repeats * self.statePerRepeat
        # solving for eigen states energy close to sigma
        self.matrixSigma = (np.min(self.xVc) + np.max(self.xVc))/2
        self.Eshift = self.periodL * self.EField * EUNIT
        self.status = 'unsolved'

    def xLayerMask(self, n: int) -> np.ndarray:
        """Return the mask for the given layer number `n`.
        A left and right extra point is included for plotting purposes."""
        xSlt = (self.xLayerNums == n)
        xSlt = (xSlt | np.roll(xSlt, 1) | np.roll(xSlt, -1))
        return ~xSlt

    def _shift_psi(self, psi, n) -> np.ndarray:
        if n == 0:
            return psi
        else:
            return np.interp(self.xPoints-self.periodL*n, self.xPoints, psi,
                             left=0, right=0)

    def _basis_shifter(self, ns: List[int], psis0: np.ndarray,
                       eigenEs0: np.ndarray, xPoints: np.ndarray = None
                       ) -> Tuple[np.ndarray, np.ndarray]:
        if xPoints is None:
            xPoints = self.xPoints
        period = self.periodL
        psis = np.empty((0, self.xPoints.size))
        eigenEs = np.empty(0)
        emax = np.max(self.xVc)
        for n in sorted(ns, reverse=True):
            psisn = np.array([np.interp(self.xPoints-period*n, xPoints, psi,
                                        left=0, right=0) for psi in psis0])
            # filter almost zero solutions)
            if len(psisn) > 0:
                es = eigenEs0 - self.Eshift * n
                idx = (np.sum(psisn**2, axis=1)*self.xres > 0.1) & (es < emax)
                psis = np.concatenate((psis, psisn[idx]))
                eigenEs = np.concatenate((eigenEs, es[idx]))
        return psis, eigenEs

    def populate_material(self):
        """This should be overridden to yield bandParams"""
        raise NotImplementedError('Material property is not implemented')

    def _reset_cache(self):
        # TODO: compatibility with shift
        self._cacheDipole = [[None]*len(self.eigenEs)
                             for _ in range(len(self.eigenEs))]
        self._cacheLO = [[None]*len(self.eigenEs)
                         for _ in range(len(self.eigenEs))]
        self.reset_IFR_cache()

    def reset_IFR_cache(self):
        """Reset cache for IFR scattering. This is exposed s.t. eigen solver
        results can be maintained after changing IFR setting."""
        self._cacheIFR = [[None]*len(self.eigenEs)
                          for _ in range(len(self.eigenEs))]
        self._cacheGamma = [[None]*len(self.eigenEs)
                            for _ in range(len(self.eigenEs))]
        if self.status.startswith('solved'):
            self.status = 'solved'

    def solve_whole(self) -> np.ndarray:
        """Solve the whole structure. Will choose from _solve_whole_ode or
        _solve_whole_matrix according to self.solver.

        Yield
        -----
        eigenEs : np.array of float
            the eigenenergy of the layer structure
        psis : np.array of float
            the wave function
        """
        if self.solver == 'ODE':
            self._solve_whole_ode()
        elif self.solver == 'matrix':
            self._solve_whole_matrix()
        else:
            raise NotImplementedError(
                'The {} solver is not implemented'.format(self.solver))
        self._reset_cache()
        self.status = 'solved'
        return self.eigenEs

    def _solve_whole_ode(self) -> np.ndarray:
        if self.crystalType == 'simple':
            self.eigenEs = onedq.cSimpleSolve1D(self.xres, self.Es,
                                                self.xVc, self.xMc)
            self.psis = onedq.cSimpleFillPsi(self.xres, self.eigenEs,
                                             self.xVc, self.xMc)
            return self.eigenEs
        xBand = onedq.PyBand(self.crystalType, *self.bandParams)
        self.eigenEs = onedq.cBandSolve1D(self.xres, self.Es, self.xVc, xBand)
        self.psis = onedq.cBandFillPsi(
            self.xres, self.eigenEs, self.xVc, xBand)

        if self.crystalType == 'ZincBlende':
            # To restore lh and so band to keep consistent with matrix solver
            xEg, xF, xEp, xESO = self.bandParams
            kunit = hbar**2/(2*e0*m0*(1E-10*self.xres)**2)
            xP = sqrt(xEp * kunit)
            dphic = np.zeros(self.psis.shape)
            dphic[:, 1:-1] = (self.psis[:, 2:] - self.psis[:, :-2])/2
            self.philh = np.zeros(self.psis.shape)
            xE = np.broadcast_to(self.eigenEs.reshape(-1, 1), self.psis.shape)
            xE = xE - self.xVc
            self.philh = -sqrt(2/3)*xP/(xE+xEg) * dphic
            self.phiso = sqrt(1/3)*xP/(xE+xEg+xESO) * dphic
        return self.eigenEs

    def _solve_whole_matrix(self) -> np.ndarray:
        # unit eV/step^2
        kunit = hbar**2/(2*e0*m0*(1E-10*self.xres)**2)
        if self.crystalType == 'simple':
            # populate mass half grid self.xMc[i] is m at i+-0.5
            layerCumSum = [0] + np.cumsum(self.layerWidths).tolist()
            periodL = layerCumSum[-1]
            self.xMcplus = np.empty(self.xPoints.size)
            self.xMcminus = np.empty(self.xPoints.size)
            for n in range(len(self.layerWidths)):
                Indices = np.logical_or.reduce([
                    (self.xPoints >= layerCumSum[n]+k*periodL-self.xres/2)
                    & (self.xPoints < layerCumSum[n+1]+k*periodL-self.xres/2)
                    for k in range(self.repeats)])
                self.xMcplus[Indices] = self.layerMc(n)
            self.xMcplus[-1] = self.xMcplus[-2]
            self.xMcminus[1:] = self.xMcplus[:-1]
            self.xMcminus[0] = self.xMcminus[1]

            # diagonal and sub-diagonal of Hamiltonian
            self.Hdiag = kunit*(1/self.xMcplus + 1/self.xMcminus) + self.xVc
            self.Hsubd = -kunit / self.xMcplus[:-1]
            # self.eigenEs, self.psis = slg.eigh_tridiagonal(
            #     self.Hdiag, self.Hsubd, select='v',
            #     select_range=(np.min(self.xVc), np.max(self.xVc)))
            self.Hsubd = -kunit / self.xMcplus
            N = len(self.xPoints)
            self.Hsparse = sparse.diags([self.Hsubd, self.Hdiag, self.Hsubd],
                                        [-1, 0, 1], shape=(N, N))
            self.eigenEs, self.psis = splg.eigsh(
                self.Hsparse, self.matrixEigenCount, sigma=self.matrixSigma,
                tol=1E-8)
            self.psis /= sqrt(self.xres)
            self.psis = self.psis.T
            return self.eigenEs
        if self.crystalType == 'ZincBlende':
            self.populate_Kane_matrix()
            # self.eigen_all, self.psi_all = slg.eig_banded(
            #     self.HBanded, select='v', select_range=(Es_low, Es_hi))
            self.eigen_all, self.psi_all = splg.eigsh(
                self.Hsparse, self.matrixEigenCount, sigma=self.matrixSigma,
                tol=1E-8)
            # filter artifacts
            idx = np.abs(self.psi_all[0, :]) < 1E-1
            self.eigen_all = self.eigen_all[idx]
            self.psi_all = self.psi_all[:, idx]
            # normalization should be sum(self.psi_all**2)*self.xres = 1
            self.psi_all /= sqrt(self.xres)
            for n in range(len(self.eigen_all)):
                phic = self.psi_all[::3, n]
                # for consistency of the phase definition with the ODE solver
                if phic[np.argmax(np.abs(phic) > 1E-3)] < 0:
                    self.psi_all[:, n] = -self.psi_all[:, n]
            self.psis = np.zeros((self.eigen_all.shape[0], self.xPoints.size))
            psis = self.psi_all[::3, :].T
            self.psis[:, :-1] = (psis[:, 1:] + psis[:, :-1])/2
            self.philh = self.psi_all[1::3, :].T
            self.phiso = self.psi_all[2::3, :].T
            self.eigenEs = self.eigen_all
            return self.eigenEs
        raise NotImplementedError('Matrix solver is not implemented '
                                  'for {}'.format(self.crystalType))

    def psi_overlap(self, upper: int, lower: int, shift=0) -> np.ndarray:
        """Return psi[upper] * psi[lower] with psi[lower] shifted by shift
        number of periods."""
        if self.crystalType == 'ZincBlende':
            return sum(
                phi[upper] * self._shift_psi(phi[lower], shift)
                for phi in (self.psis, self.philh, self.phiso))
        # default fallback and crystalType == 'simple'
        return self.psis[upper] * self._shift_psi(self.psis[lower], shift)

    def populate_Kane_matrix(self) -> sparse.spmatrix:
        """
        Populate the finite difference Hamiltonian operation for the Kane
        3 band model.

        Return
        ------
        Hsparse : scipy.sparse.spmatrix
            A sparse matrix for the Hamiltonian

        Yield
        -----
        Hsparse : scipy.sparse.spmatrix
            Same above
        Hbanded : np.ndarray
            Upper banded form of the matrix for the banded solver, as
            used in scipy.linalg.eig_banded
        """
        assert self.crystalType == 'ZincBlende'
        kunit = hbar**2/(2*e0*m0*(1E-10*self.xres)**2)
        N = len(self.xPoints)
        xEg, xF, xEp, xESO = self.bandParams
        xFhalf = np.empty(N)
        xFhalf[1:] = (xF[1:] + xF[:-1])/2
        xFhalf[0] = xF[0]
        xVcHalf = np.empty(N)
        xVcHalf[1:] = (self.xVc[1:] + self.xVc[:-1])/2
        xVcHalf[0] = self.xVc[0]
        self.HBanded = np.zeros((4, 3*N))
        self.HBanded[3, ::3] = 2*(1 + 2*xFhalf)*kunit + xVcHalf
        self.HBanded[3, 1::3] = self.xVc - xEg  # lh band
        self.HBanded[3, 2::3] = self.xVc - xEg - xESO  # so band
        P = sqrt(xEp * kunit)
        self.HBanded[2, 1::3] = sqrt(2/3)*P
        self.HBanded[2, 3::3] = sqrt(1/3)*P[:-1]
        self.HBanded[1, 2::3] = -sqrt(1/3)*P
        self.HBanded[1, 3::3] = -sqrt(2/3)*P[:-1]
        self.HBanded[0, 3::3] = -(1 + 2*xF[:-1]) * kunit
        if hasattr(self, 'luttinger'):
            gamma1, gamma2, _ = self.luttinger
            tlh = gamma1 + 2*gamma2 - 2*xEp/xEg/3
            tso = gamma1 - xEp/xEg/3
            tlhhalf = (tlh[1:] + tlh[:-1])/2
            tsohalf = (tso[1:] + tso[:-1])/2
            tlh[1:-1] = (tlhhalf[1:] + tlhhalf[:-1])/2
            tso[1:-1] = (tsohalf[1:] + tsohalf[:-1])/2
            self.HBanded[3, 1::3] -= 2*tlh*kunit
            self.HBanded[3, 2::3] -= 2*tso*kunit
            self.HBanded[0, 4::3] = tlhhalf*kunit
            self.HBanded[0, 5::3] = tsohalf*kunit
        self.Hsparse = sparse.diags(
            [self.HBanded[0, 3:], self.HBanded[1, 2:], self.HBanded[2, 1:],
             self.HBanded[3, :],
             self.HBanded[2, 1:], self.HBanded[1, 2:], self.HBanded[0, 3:]
             ], [-3, -2, -1, 0, 1, 2, 3], shape=(3*N, 3*N))
        return self.Hsparse

    def _isBasisBreak(self, n: int) -> bool:
        if self.basisInjectorAR:
            if not self.layerARs[n-1] and self.layerARs[n]:
                return True
        if self.basisARInjector:
            if self.layerARs[n-1] and not self.layerARs[n]:
                return True
        return False

    def solve_basis(self) -> np.ndarray:
        """
        solve basis for the QC device, with each basis being the eigen mode of
        a separate part of the layer structure

        Yield
        -----
        eigenEs : np.array of float
            the eigenenergy of the layer structure
        psis : np.array of float
            the wave function
        """
        startIdx = []
        endIdx = []
        # Get the region of interest
        if self.basisAROnly:
            if self.layerARs[0]:
                startIdx.append(0)
            for n in range(1, len(self.layerARs)):
                if not self.layerARs[n-1] and self.layerARs[n]:
                    startIdx.append(n)
                if self.layerARs[n-1] and not self.layerARs[n]:
                    endIdx.append(n)
            if self.layerARs[-1]:
                endIdx.append(len(self.layerARs))
        else:
            startIdx.append(-1)
            for n in range(1, len(self.layerARs)):
                if self._isBasisBreak(n):
                    barrier = n if self.layerVc(n) > self.layerVc(n-1) else n-1
                    if barrier != startIdx[-1]:
                        startIdx.append(barrier)
                        endIdx.append(barrier + 1)
            if self._isBasisBreak(0):
                barrier = 0 if self.layerVc(0) > self.layerVc(-1) else -1
                barrier += len(self.layerWidths)
                startIdx.append(barrier)
                endIdx.append(barrier + 1)
            if len(endIdx) == 0:
                startIdx = [0]
                endIdx = [len(self.layerWidths)]
            else:
                startIdx = startIdx[1:]
                endIdx = endIdx[1:] + [endIdx[0] + len(self.layerWidths)]

        self.eigenEs = np.empty((0))
        self.psis = np.empty((0, self.xPoints.size))
        for n in range(0, len(startIdx)):
            dCL = copy.copy(self)
            dCL._reset_for_basis(startIdx[n], endIdx[n])
            dCL.populate_x()
            dCL.solve_whole()

            # map dCL result back to self
            shift = sum(self.layerWidths[:startIdx[n]]) - BASISPAD
            psis, eigenEs = self._basis_shifter(
                range(-1, self.repeats), dCL.psis,
                dCL.eigenEs - shift*self.EField*EUNIT,
                dCL.xPoints + shift)
            self.eigenEs = np.concatenate((self.eigenEs, eigenEs))
            self.psis = np.concatenate((self.psis, psis))
        self._reset_cache()
        self.status = 'basis'
        return self.eigenEs

    def _reset_for_basis(self, start: int, end: int):
        """Reset the parameters for only solving the layers of [start:end].
        This is a helper method for solve_basis"""
        self.repeats = 1
        self.statePerRepeat = (
            self.statePerRepeat * (end - start) // len(self.layerWidths))
        self.layerWidths = (self.layerWidths*2)[start:end]
        self.layerWidths[0] += BASISPAD
        self.layerWidths[-1] += BASISPAD
        self.layerARs = (self.layerARs*2)[start:end]
        self._layerVc = (self._layerVc*2)[start:end]
        self._layerMc = (self._layerMc*2)[start:end]

    def _dipole(self, upper: int, lower: int, shift: int = 0) -> float:
        """Return Electrical dipole between upper and lower states
        in unit angstrom, update self.dipole.
        shift means lower state is shifted by this number of period.
        Should be called for any other related physics quantities."""
        # TODO: clean up self cache for upper -> lower states
        if self.solver == 'ODE':
            psi_u = self.psis[upper, :]
            psi_l = self._shift_psi(self.psis[lower, :], shift)
            Eu = self.eigenEs[upper]
            El = self.eigenEs[lower] - shift * self.Eshift
            xInvMc_u = self._xBandMassInv(Eu)
            xInvMc_l = self._xBandMassInv(El)
            # Eq.(8) in PhysRevB.50.8663, with more precise mass
            # m_j comes from the density of states of transition final state
            z = np.trapezoid(psi_u * xInvMc_l * np.gradient(psi_l) -
                             np.gradient(psi_u) * xInvMc_u * psi_l)
            z *= hbar**2 / (2*(El-Eu)*e0*m0) / (1E-10)**2  # Angstrom
        elif self.solver == 'matrix' and self.crystalType == 'ZincBlende':
            # In principle this and above is equivalent, and numerically
            # approximately the same as tested in unit tests.
            # The above is remained for legacy reasons.
            z = self.xres * np.trapezoid(
                self.psi_overlap(upper, lower, shift) * self.xPoints)
        else:
            raise NotImplementedError(
                '{} not implemented for dipole'.format(self.solver))
        return z

    def dipole(self, upper, lower) -> float:
        if self._cacheDipole[upper][lower] is None:
            self._cacheDipole[upper][lower] = self._dipole(upper, lower)
        return self._cacheDipole[upper][lower]

    def _lo_transition(self, upper: int, lower: int, shift: int = 0) -> float:
        # TODO: finite temperature version
        Eu = self.eigenEs[upper]
        El = self.eigenEs[lower] - shift * self.Eshift
        if Eu - El - self.avghwLO < 0:
            # energy difference is smaller than a LO phonon
            # LO phonon scattering doesn't happen
            return INV_INF
        psi_l_sq = self.psi_overlap(lower, lower)
        ml = m0 * np.trapezoid(self.xMc * psi_l_sq) * self.xres
        kl = sqrt(2 * ml / hbar**2 * (Eu - El - self.avghwLO) * e0)
        N = self.xPoints.size
        if onedq is None:
            convpsi = fft.irfft(np.abs(fft.rfft(
                self.psi_overlap(upper, lower, shift), 2*N))**2)[:N]
            Iij = 2*self.xres**2*np.trapezoid(
                exp(-kl*self.xPoints*1E-10)*convpsi)
        # C implementation
        else:
            Iij = onedq.cLOphononScatter(
                self.xres, kl, self.psi_overlap(upper, lower, shift))
        return (ml * e0**2 * self.avghwLO * e0 / hbar * Iij
                / (4 * hbar**2 * self.epsrho * eps0 * kl)) / 1e12  # unit ps^-1

    def lo_transition(self, upper: int, lower: int) -> float:
        """The LO phonon transition lifetime from upper to lower,
        at zero temperature.
        This is using cached results.
        if status is 'solved-full' and the state is a recognized state of a
        period, it's calculated via translation of the wavefunction, otherwise
        it's calculated based on row self.psis
        """
        if self.status == 'solved-full':
            try:
                pu, ushift = self.periodMap[upper]
                pl, lshift = self.periodMap[lower]
                if lshift - ushift not in (0, 1, -1):
                    return INV_INF
                return self._pLO[lshift-ushift][pl][pu]
            except (TypeError, AttributeError):
                # TypeError is when self.periodMap return None
                # AttributeError is when self.periodMap does not exist
                pass
        if self._cacheLO[upper][lower] is None:
            self._cacheLO[upper][lower] = self._lo_transition(
                upper, lower)
        return self._cacheLO[upper][lower]

    def lo_lifetime(self, state: int) -> float:
        """ Return the life time due to LO phonon scattering of the
        given state(label)
        This is using cached results.
        if status is 'solved-full' and the state is a recognized state of a
        period, it's calculated via translation of the wavefunction, otherwise
        it's calculated based on row self.psis
        """
        if self.status == 'solved-full':
            try:
                return 1/sum(np.sum(self._pLO[n][:, self.periodMap[state][0]])
                             for n in range(3))
            except (TypeError, AttributeError):
                # TypeError is when self.periodMap return None
                # AttributeError is when self.periodMap does not exist
                pass
        Ei = self.eigenEs[state]
        if onedq is None:
            return 1/sum(self.lo_transition(state, q) for q in range(state)
                         if self.eigenEs[q] <= Ei - self.avghwLO)
        idxs = self.eigenEs <= Ei - self.avghwLO
        Ejs = self.eigenEs[idxs]
        idxs, = idxs.nonzero()
        psi_js_sq = np.array([self.psi_overlap(idx, idx) for idx in idxs])
        mjs = m0 * np.trapezoid(self.xMc * psi_js_sq, axis=1) * self.xres
        kls = sqrt(2 * mjs / hbar**2 * (Ei - Ejs - self.avghwLO) * e0)
        fjs = (mjs * e0**2 * self.avghwLO * e0 / hbar
               / (4 * hbar**2 * self.epsrho * eps0 * kls))
        psi_ijs = np.array([self.psi_overlap(state, idx) for idx in idxs])
        Iijtotal = onedq.cLOtotal(self.xres, kls, psi_ijs, fjs)
        return 1e12 / Iijtotal if Iijtotal > 0 else 1E20

    def _ifr_transition(self, upper: int, lower: int, shift: int = 0
                        ) -> Tuple[float, float]:
        # TODO: finite temperature
        psi_usq = self.psi_overlap(upper, upper)
        psi_lsq = self.psi_overlap(lower, lower)
        psi_ul = self.psi_overlap(upper, lower, shift)
        Eu = self.eigenEs[upper]
        El = self.eigenEs[lower] - shift * self.Eshift
        if Eu < El:
            return INV_INF, -1
        mu = m0 * np.trapezoid(self.xMc * psi_usq) * self.xres
        ml = m0 * np.trapezoid(self.xMc * psi_lsq) * self.xres
        kl = sqrt(2 * ml / hbar**2 * (Eu-El) * e0)
        tauInv = 0
        gamma = 0
        zn = 0
        layerN = len(self.layerWidths)
        for _ in range(self.repeats):
            for n in range(layerN):
                lamb = self.ifrLambda[n] * 1E-10  # to m
                delt = self.ifrDelta[n]
                dU = (self.layerVc((n+1) % layerN) - self.layerVc(n))
                dU *= e0  # to J
                # find interface
                zn += self.layerWidths[n]
                zIdx = np.argmax(self.xPoints >= zn)
                if zIdx == 0 or zIdx == len(self.xPoints)-1:
                    continue
                z1 = self.xPoints[zIdx-1]
                z2 = self.xPoints[zIdx]

                def interpZ(psi):
                    return (psi[zIdx-1]*(zn-z2) - psi[zIdx]*(zn-z1))/(z2-z1)
                psi_usqz = interpZ(psi_usq)
                psi_lsqz = interpZ(psi_lsq)
                psi_ulz = interpZ(psi_ul)
                scale = pi / hbar**3 * delt**2 * lamb**2 * dU**2
                tauInv += scale * ml * psi_ulz**2 * exp(-lamb**2 * kl**2 / 4)
                gamma += (scale * (psi_usqz - psi_lsqz)
                          * (psi_usqz * mu - psi_lsqz * ml))/2
        return tauInv/1E12, gamma/1E12  # unit ps^-1

    def ifr_transition(self, upper: int, lower: int) -> float:
        r"""Calculate the interface roughness (IFR) transition rate from
        upper to lower state at zero temperature, in unit ps^-1.

        .. math::
            \frac{1}{\tau_{ij}^\text{IFR}} =
            \frac{\pi m^*_j}{\hbar^3} \sum_n
            \Delta_n^2\Lambda_n^2\delta U_n^2
            \left|\psi_i(z_n)\psi_j^*(z_n)\right|^2
            \mathrm e^{- \Lambda^2 m_j^* (E_i - E_j))/2\hbar^2}

        This is using cached results.
        if status is 'solved-full' and the state is a recognized state of a
        period, it's calculated via translation of the wavefunction, otherwise
        it's calculated based on row self.psis
        """
        assert self.includeIFR
        if self.status == 'solved-full':
            try:
                pu, ushift = self.periodMap[upper]
                pl, lshift = self.periodMap[lower]
                if lshift - ushift not in (0, 1, -1):
                    return INV_INF
                return self._pIFR[lshift-ushift][pl][pu]
            except (TypeError, AttributeError):
                # TypeError is when self.periodMap return None
                # AttributeError is when self.periodMap does not exist
                pass
        if self._cacheIFR[upper][lower] is not None:
            return self._cacheIFR[upper][lower]
        tauInv, gamma = self._ifr_transition(upper, lower)
        self._cacheIFR[upper][lower] = tauInv
        if gamma > 0:
            self._cacheGamma[upper][lower] = gamma
            self._cacheGamma[lower][upper] = gamma
        return self._cacheIFR[upper][lower]

    def ifr_lifetime(self, state: int) -> float:
        """Return to total life time due to IFR scattering.
        This is using cached results.
        if status is 'solved-full' and the state is a recognized state of a
        period, it's calculated via translation of the wavefunction, otherwise
        it's calculated based on row self.psis
        """
        assert self.includeIFR
        if self.status == 'solved-full':
            try:
                return 1/sum(np.sum(self._pIFR[n][:, self.periodMap[state][0]])
                             for n in range(3))
            except (TypeError, AttributeError):
                # TypeError is when self.periodMap return None
                # AttributeError is when self.periodMap does not exist
                pass
        return 1/sum(self.ifr_transition(state, q) for q in range(state))

    def lifetime(self, state: int) -> float:
        """A convenience wrap of return total lifetime of LO and IFR scattering
        or only LO scattering depending on self.includeIFR."""
        if self.status == 'solved-full':
            try:
                return 1/self.decayRates[self.periodMap[state][0]]
            except (TypeError, AttributeError):
                # TypeError is when self.periodMap return None
                # AttributeError is when self.periodMap does not exist
                pass
        if self.includeIFR:
            return 1/(1/self.ifr_lifetime(state) + 1/self.lo_lifetime(state))
        else:
            return self.lo_lifetime(state)

    def ifr_broadening(self, upper: int, lower: int) -> float:
        """Interface roughness induced broadening"""
        # TODO: shift life time
        if self.status == 'solved-full':
            try:
                pu, ushift = self.periodMap[upper]
                pl, lshift = self.periodMap[lower]
                if lshift - ushift not in (0, 1, -1):
                    return 1/INV_INF
                return self._pGamma[lshift-ushift][pl][pu]
            except (TypeError, AttributeError):
                # TypeError is when self.periodMap return None
                # AttributeError is when self.periodMap does not exist
                pass
        if self._cacheGamma[upper][lower] is None:
            self.ifr_transition(upper, lower)
            self.ifr_transition(lower, upper)
        return self._cacheGamma[upper][lower]

    def dephasing(self, upper: int, lower: int) -> float:
        r"""Calculate the broadening gamma of transition between upper ->
        lower transition, return gamma in unit eV as in Lorentzian:

        .. math::
            \mathcal L(\omega) =
            \frac{1}{\pi} \frac{\gamma}{\gamma^2 + (\omega - \omega_0)^2}

        If IFR scattering is included the broadening is calculated dominantly
        from IFR broadening and finite lifetime of upper and lower states.
        Otherwise 0.1 is returned.
        """
        if not self.includeIFR:
            Eu = self.eigenEs[upper]
            El = self.eigenEs[lower]
            de = np.abs(Eu - El)
            return 0.05 * de
        gamma_u = 1/self.ifr_lifetime(upper) + 1/self.lo_lifetime(upper)
        gamma_l = 1/self.ifr_lifetime(lower) + 1/self.lo_lifetime(lower)
        gamma_parallel = self.ifr_broadening(upper, lower)
        # 1E12: ps^-1 -> Hz
        return (gamma_parallel + (gamma_u + gamma_l)/2) * 1E12 * hbar / e0

    def period_recognize(self, tol: float = 5E-5) -> np.ndarray:
        """Pick a set of eigen states as states in a period.

        Return
        ------
        singlePeriodIdx : np.array of int
            These are indices for the recognized states of a single period.

        Yield
        -----
        unBound : set of int
            includes index of states that are not well bounded.
        """
        # TODO: try look from the low energy side
        periodIdx = self.periodL / self.xres
        psisq = np.abs(self.psis)**2
        self.starts = np.argmax(psisq > tol, axis=1)
        self.ends = np.argmax(psisq[:, ::-1] > tol, axis=1)
        self.periodIdx = np.arange(len(self.eigenEs))[
            (self.starts > periodIdx/3) & (self.starts < 4*periodIdx/3)]
        # check if all states are far away enough from boundary
        self.unBound = set()
        barrierBound = np.max(self.xVc + self.xVField) - self.xVField
        for n in self.periodIdx:
            if self.ends[n] < periodIdx / 3:
                self.unBound.add(n)
                if self.eigenEs[n] < barrierBound[-self.ends[n]]:
                    # TODO: use warning package
                    print('State No.{} is close to the right boundary. '
                          'More repeats may be needed'.format(n))
        # print(self.singlePeriodIdx, self.looselyBounded)
        if len(self.periodIdx) - len(self.unBound) == 0:
            raise StateRecognizeError('No well bounded state found. '
                                      'Try increase repeats.')
        return self.periodIdx

    def period_map_build(self, tol: float = 3E-2, etol: float = 1E-3
                         ) -> List[Tuple[int, int]]:
        """Map states to self.singlePeriodIdx, self.periodMap[n] is a tuple of
        (state index in self.singlePeriodIdx, shift of period(s) or
        None meaning it's not mapped."""
        self.periodMap = [None] * len(self.eigenEs)
        singleE = self.eigenEs[self.periodIdx]
        for n, state in enumerate(self.periodIdx):
            self.periodMap[state] = (n, 0)
        for state, en in enumerate(self.eigenEs):
            if self.periodMap[state] is not None:
                continue
            psi = self.psis[state]
            for shift in range(1, self.repeats):
                en_shifted = en + shift * self.Eshift
                psi_shifted = self._shift_psi(psi, -shift)
                for n, sE in enumerate(singleE):
                    if sE < en_shifted - etol:
                        continue
                    if sE > en_shifted + etol:
                        break
                    sState = self.periodIdx[n]
                    wfDiff = np.trapezoid((psi_shifted - self.psis[sState])**2)
                    wfDiff *= self.xres
                    if wfDiff < tol:
                        # effective an L2 norm here, tested better than
                        # L1 or L-max norm
                        self.periodMap[state] = (n, shift)
                        break
                if self.periodMap[n] is not None:
                    break
        return self.periodMap

    def full_population(self) -> np.ndarray:
        """Calculate the electron full population on states, assuming the
        result of solve_whole and periodRecognize is valid and no state has
        coupling with states two or more periods away.

        Return
        ------
        population : np.array of float, dim = len(periodIdx)
            The population of electrons in the first recognized period, state
            label self.PeriodIdx[n]

        Yield
        -----
        transitions : np.array of float, dim = len(periodIdx)*len(periodIdx)
            The transition rate from self.PeriodIdx[i] to self.PeriodIdx[j]
        decayRates : np.array of float, dim = len(periodIdx)
            Inverse of lifetimes
        flow : float
            The flow of carrier, in unit ps^-1 (carrier density normalize to 1)
        """
        assert self.status.startswith('solved')
        idxPeriod = len(self.periodIdx)
        # p for cache for the periodic version
        self._pLO = [np.zeros((idxPeriod, idxPeriod)) for _ in range(3)]
        self._pIFR = [np.zeros((idxPeriod, idxPeriod)) for _ in range(3)]
        self._pGamma = [np.zeros((idxPeriod, idxPeriod)) for _ in range(3)]
        for i, lower in enumerate(self.periodIdx):
            for j, upper in enumerate(self.periodIdx):
                # keep this consistent with loMatrix etc
                for s in ((1, -1) if lower == upper else (1, 0, -1)):
                    self._pLO[s][j][i] = self._lo_transition(lower, upper, s)
                    self._pIFR[s][j][i], gamma = (
                        self._ifr_transition(lower, upper, s))
                    if gamma > 0:
                        self._pGamma[s][j][i] = self._pGamma[s][i][j] = gamma
        forward = self._pLO[1] + self._pIFR[1]
        backward = self._pLO[-1] + self._pIFR[-1]
        internal = self._pLO[0] + self._pIFR[0]
        self.transitions = internal + forward + backward
        self.decayRates = np.sum(self.transitions, axis=0)
        self.population = null_space(
            np.diag(-self.decayRates) + self.transitions)
        if self.population.shape[1] != 1:
            raise ValueError('More than one steady state found. ',
                             self.population.shape[1])
        self.population = self.population[:, 0].T
        self.population /= np.sum(self.population)
        if self.carrierLeak > 5E-2:
            # TODO: use warning package
            print("The structure seems highly leak or more period needed.")
            print("    Most leaking (%.2f%%) unbounded state: %d" %
                  max((self.population[n], state)
                      for n, state in enumerate(self.periodIdx)
                      if state in self.unBound))
        self.flow = np.sum((forward - backward) @ self.population)
        # print("transition: ", self.transitions)
        # print("lifetimes: ", -1/transfer.diagonal())
        # print("population: ", self.population)
        self.status = 'solved-full'
        return self.population

    @property
    def carrierLeak(self) -> float:
        return sum(
            self.population[n] for n, state in enumerate(self.periodIdx)
            if state in self.unBound)

    def state_population(self, state: int) -> float:
        """This method is only valid after fullPopulation has been called"""
        if self.periodMap[state] is None:
            return None
        return self.population[self.periodMap[state][0]]

    def _xBandMassInv(self, energy: float) -> np.ndarray:
        if self.crystalType == 'simple':
            return self.xMc
        if self.crystalType == 'ZincBlende':
            xEg, xF, xEp, xESO = self.bandParams
            E = energy - self.xVc
            return 1 + 2*xF + 1/3 * xEp/(E+xEg+xESO) + 2/3 * xEp/(E+xEg)
        else:
            raise NotImplementedError(
                 'Material property for {} crystal is not implemented'.format(
                     self.crystalType
                 ))


class QCLayers(SchrodingerLayer):
    r"""Class for Quantum Cascade Layers

Parameters
----------
substrate : str
    The substrate material for the device, which determines the well and
    barrier material

    ========= ================================ ================================
    substrate              well                         barrier
    ========= ================================ ================================
    InP       In\ :sub:`x`\ Ga\ :sub:`1-x`\ As Al\ :sub:`1-x`\ In\ :sub:`x`\ As
    GaAs      Al\ :sub:`x`\ Ga\ :sub:`1-x`\ As Al\ :sub:`x`\ Ga\ :sub:`1-x`\ As
    GaSb      InAs\ :sub:`y`\ Sb\ :sub:`1-y`   Al\ :sub:`x`\ Ga\ :sub:`1-x`\ Sb
    ========= ================================ ================================

materials :
    Name of alloys for the heterostructure materials, len >= 2
moleFracs :
    mole fraction for each possible layer material, len = Mp. of materials
xres :
    Position resolution, in Armstrong
Eres :
    Energy resolution, in meV.
    This number being too large may results in miss of some states while
    this being too small will make a long computation time.
    The parameter does not mean the accuracy of the eigen energy. It's
    required for algorithm reasons because of lack of a universal global
    root finding.
statePerRepeat :
    Number of states per repeat, used for calculating matrixEigenCount
wl :
    The wavelength for the design, in unit um, gain spectrum and optimization,
    but doesn't go into quantum solver

layerWidths :
    Width of each layer, in angstrom. len = No. of layers
layerMtrls :
    Label of materials, depending on substrate. len = No. of layers
layerDopings :
    Doping per volume in unit 1e17 cm-3. len = No. of layers
layerARs :
    Binaries indicating if the layer is active(True) or not(False),
    only affects basis solver. len = No. of layers

customIFR :
    Wether to use a customized IFR parameter rather than a material determined
    parameter.
mtrlIFRLambda :
    The interface roughness lambda after materials[n], len = No. of materials
mtrlIFRDelta :
    The interface roughness delta after materials[n], len = No. of materials

EField :
    External (static) electrical field, in kV/cm = 1e5 V/m
repeats :
    Number of repeat times for the given structure
T :
    Temperature of the device, affecting material property

basisAROnly :
    For basis solver if only the Active Region (AR) should be solved.
basisARInjector :
    For basis solver if there should be separator between AR->Injector
basisInjectorAR :
    For basis solver if there should be separator between Injector->AR

solver :
    The solver used for the eigen problem: 'ODE' or 'matrix'.
    By default 'ODE' if C library exists, 'matrix' is a full back.
includeIFR :
    Weather to include IFR scattering for performance estimation.
matrixEigenCount :
    The number of eigen pairs to calculate in the 'matrix' solver.
    It would be very expensive to calculate all of them.

status :
    - 'unsolved' meaning the structure is not solved yet.
    - 'basis' meaning the eigen problem is solved for basis
    - 'solved' meaning the eigen problem is solved.
    - 'solved-full' meaning the population distribution is known.

description :
    Description of the data. For book-keeping purposes.
    """
    materials: List[str]
    moleFracs: List[float]
    wl: float
    layerMtrls: List[int]
    layerDopings: List[float]
    customIFR: bool
    mtrlIFRLambda: List[float]
    mtrlIFRDelta: List[float]
    T: float
    description: str

    def __init__(self, substrate="InP", materials=["InGaAs", "AlInAs"],
                 moleFracs=[0.53, 0.52], xres=0.5, Eres=0.5, statePerRepeat=20,
                 layerWidths=[10.0], layerMtrls=None, layerDopings=None,
                 customIFR=False, mtrlIFRLambda=None, mtrlIFRDelta=None,
                 ifrDelta=None, ifrLambda=None,
                 layerARs=None, EField=0, repeats=3, T=300.0, solver="ODE",
                 description="", wl=3.0):
        assert isinstance(layerWidths, list)
        assert isinstance(materials, list)
        assert isinstance(moleFracs, list)
        N = len(layerWidths)
        M = len(materials)
        assert M >= 1
        assert len(moleFracs) == M
        self.substrate = substrate
        self.materials = materials
        self.moleFracs = moleFracs
        self.layerMtrls = [0]*N if layerMtrls is None else layerMtrls
        self.layerDopings = [0.0]*N if layerDopings is None else layerDopings
        self.temperature = T
        self.customIFR = customIFR
        if not customIFR:
            if isinstance(mtrlIFRDelta, list):
                assert len(mtrlIFRDelta) == M
                assert isinstance(mtrlIFRLambda, list)
                assert len(mtrlIFRLambda) == M
                self.mtrlIFRDelta = mtrlIFRDelta
                self.mtrlIFRLambda = mtrlIFRLambda
            else:
                self.mtrlIFRDelta = [mtrlIFRDelta or 0.0] * M
                self.mtrlIFRLambda = [mtrlIFRLambda or 0.0] * M
            ifrDelta, ifrLambda = self._get_IFRList()
        self.description = description
        super().__init__(xres=xres, Eres=Eres, statePerRepeat=statePerRepeat,
                         layerWidths=layerWidths, layerARs=layerARs,
                         ifrDelta=ifrDelta, ifrLambda=ifrLambda,
                         EField=EField, repeats=repeats)
        self.crystalType = Material.MParam[substrate]["Crystal"]
        self.subM = Material.Material(self.substrate, self.temperature)
        self.wl = wl
        self.solver = solver
        if onedq is None:
            self.solver = 'matrix'
        self.update_mtrls()

    def __copy__(self):
        return QCLayers(
            substrate=self.substrate, materials=copy.copy(self.materials),
            moleFracs=copy.copy(self.moleFracs), xres=self.xres,
            Eres=self.Eres, statePerRepeat=self.statePerRepeat,
            layerWidths=copy.copy(self.layerWidths),
            layerMtrls=copy.copy(self.layerMtrls),
            layerDopings=copy.copy(self.layerDopings),
            customIFR=self.customIFR,
            mtrlIFRLambda=copy.copy(self.mtrlIFRLambda),
            mtrlIFRDelta=copy.copy(self.mtrlIFRDelta),
            ifrDelta=copy.copy(self.ifrDelta),
            ifrLambda=copy.copy(self.ifrLambda),
            layerARs=copy.copy(self.layerARs),
            EField=self.EField, repeats=self.repeats, T=self.temperature,
            solver=self.solver, description=self.description, wl=self.wl
            )

    def _get_IFRList(self) -> Tuple[List[float], List[float]]:
        """Get IFR parameters for SchrodingerLayer. Should be called
        every time the material list changes."""
        assert not self.customIFR
        if self.mtrlIFRDelta is not None:
            ifrDelta = [self.mtrlIFRDelta[m] for m in self.layerMtrls]
        else:
            ifrDelta = [0.0] * len(self.layerMtrls)
        if self.mtrlIFRLambda is not None:
            ifrLambda = [self.mtrlIFRLambda[m] for m in self.layerMtrls]
        else:
            ifrLambda = [0.0] * len(self.layerMtrls)
        return ifrDelta, ifrLambda

    def update_mtrls(self):
        """Update properties for the materials.
        This should be called every time the material parameters are changed
        directly via `materials`, `moleFracs` or `temperature`.

        Yield
        -----
        a_parallel : float
            The parallel crystal constant of the structure, determined by the
            substrate material.

        mtrlAlloys : List[Material.Alloy]
            list of the Alloy objects for processing material properties.
        """
        self.a_parallel = self.subM.param['alc']
        self.mtrlAlloys = [Material.Alloy(self.materials[idx],
                                          self.moleFracs[idx],
                                          self.temperature)
                           for idx in range(len(self.materials))]
        for al in self.mtrlAlloys:
            al.set_strain(self.a_parallel)

    def set_mtrl(self, n: int, mtrl: str = None, moleFrac: float = None):
        """Set material[n] to new material (mtrl) and/or moleFrac"""
        if mtrl is None and moleFrac is None:
            raise Exception("Nothing changed")
        if mtrl is None:
            mtrl = self.materials[n]
        if moleFrac is None:
            moleFrac = self.moleFracs[n]
        self.moleFracs[n] = moleFrac
        self.materials[n] = mtrl
        self.mtrlAlloys[n] = Material.Alloy(mtrl, moleFrac, self.temperature)
        self.mtrlAlloys[n].set_strain(self.a_parallel)

    def add_mtrl(self, mtrl: str = None, moleFrac: float = None,
                 IFRLambda: float = None, IFRDelta: float = None):
        """Add a new material possibility"""
        self.materials.append(mtrl if mtrl else
                              QCMaterial[self.substrate][0])
        self.moleFracs.append(moleFrac if moleFrac else 0.0)
        self.mtrlIFRLambda.append(IFRLambda if IFRLambda else
                                  self.mtrlIFRLambda[-1])
        self.mtrlIFRDelta.append(IFRDelta if IFRDelta else
                                 self.mtrlIFRDelta[-1])
        self.mtrlAlloys.append(Material.Alloy(
            self.materials[-1], self.moleFracs[-1], self.temperature))
        self.mtrlAlloys[-1].set_strain(self.a_parallel)

    def del_mtrl(self, n: int):
        """Delete materials labeled n.
        All layers of this material will become previous
        `materials[n-1 if n >0 else 1]`.  There should be at least two
        materials otherwise there will be error."""
        if len(self.materials) <= 2:
            raise ValueError("There should be at least 2 materials")
        self.materials.pop(n)
        self.moleFracs.pop(n)
        for i in range(len(self.layerMtrls)):
            if self.layerMtrls[i] >= n:
                self.layerMtrls[i] = (self.layerMtrls[i] - 1
                                      if self.layerMtrls[i] > 0
                                      else 0)

    def add_layer(self, n: int, width: int, mtrlIdx: int,
                  AR: bool, doping: float):
        self.layerMtrls.insert(n, mtrlIdx)
        self.layerDopings.insert(n, doping)
        super().add_layer(n, width, AR=AR)
        if not self.customIFR:
            self.ifrDelta, self.ifrLambda = self._get_IFRList()

    def rotate_layer(self):
        super().rotate_layer()
        for layerList in (self.layerMtrls, self.layerDopings):
            layerList.insert(0, layerList.pop())

    def del_layer(self, n: int):
        super().del_layer(n)
        for layerList in (self.layerMtrls, self.layerDopings):
            layerList.pop(n)

    def invert_layer(self):
        super().invert_layer()
        self.layerMtrls = self.layerMtrls[::-1]
        self.layerDopings = self.layerDopings[::-1]
        if not self.customIFR:
            self.ifrDelta, self.ifrLambda = self._get_IFRList()

    def set_substrate(self, subs: str):
        if subs in QCMaterial:
            self.substrate = subs
            self.crystalType = Material.MParam[subs]["Crystal"]
            matlN = len(self.materials)
            self.materials = (QCMaterial[subs]*matlN)[0:matlN]
            self.update_mtrls()
        else:
            raise TypeError("Substrate %s not supported" % subs)

    def set_temperature(self, T: float):
        self.temperature = T
        self.subM.set_temperature(T)
        self.update_mtrls()

    @property
    def mtrlOffset(self) -> float:
        """Return the conduction band offset (difference between highest
        conduction band and lowest conduction band energy) of materials,
        in unit eV"""
        ecgs = [alloy.param['EcG'] for alloy in self.mtrlAlloys]
        return max(ecgs) - min(ecgs)

    @property
    def netStrain(self) -> float:
        """Return average strain perpendicular to the layer plane, in
        percentage."""
        if sum(self.layerWidths) <= 1e-5:
            return -1
        totalStrain = sum(self.mtrlAlloys[self.layerMtrls[n]].eps_perp
                          * self.layerWidths[n]
                          for n in range(len(self.layerWidths)))
        return 100 * totalStrain / sum(self.layerWidths)

    def layerVc(self, n: int) -> float:
        return self.mtrlAlloys[self.layerMtrls[n]].param['EcG']

    def layerMc(self, n: int) -> float:
        return self.mtrlAlloys[self.layerMtrls[n]].param['me0']

    def populate_material(self):
        """
        Update following band structure parameters (with type *np.array
        of float*): **xVc, xVX, xVL, xVLH, xVSO,
        bandParams = (xEg, xF, xEp, xESO)**
        """
        if self.crystalType == 'ZincBlende':
            N = self.xPoints.size
            self.xDopings = np.empty(N)

            self.xVX = np.empty(N)   # X point in the band
            self.xVL = np.empty(N)   # L point in the band
            self.xVLH = np.empty(N)  # The light hole valence band
            self.xVSO = np.empty(N)  # The heavy hole valence band

            # band parameters
            xF = np.empty(N)
            xEg = np.empty(N)
            xESO = np.empty(N)
            xEp = np.empty(N)

            for n in range(len(self.layerWidths)):
                indices = (self.xLayerNums == n)
                self.xDopings[indices] = self.layerDopings[n]
                for (p, key) in ((self.xVLH, 'EvLH'), (self.xVSO, 'EvSO'),
                                 (self.xVX, 'EcX'), (self.xVL, 'EcL'),
                                 (xEg, 'EgLH'), (xESO, 'ESO'),
                                 (xEp, 'Ep'), (xF, 'F')):
                    p[indices] = self.mtrlAlloys[self.layerMtrls[n]].param[key]
            self.bandParams = (xEg, xF, xEp, xESO)

            ExtField = self.xPoints * self.EField * EUNIT
            for p in (self.xVX, self.xVL, self.xVLH, self.xVSO):
                p -= ExtField
        else:
            raise NotImplementedError(
                 'Material property for {} crystal is not implemented'.format(
                     self.crystalType
                 ))
        # LO phonon
        if sum(self.layerWidths) <= 1e-5:
            self.avghwLO = -1
            self.epsrho = 69.0
        else:
            sumhwlo = sum(self.mtrlAlloys[self.layerMtrls[n]].param['hwLO']
                          * self.layerWidths[n]
                          for n in range(len(self.layerWidths)))
            self.avghwLO = sumhwlo / sum(self.layerWidths)
            epsInf = np.array([a.param["epsInf"] for a in self.mtrlAlloys])
            epss = np.array([a.param["epss"] for a in self.mtrlAlloys])
            epsrho = 1 / (1/epsInf - 1/epss)
            self.epsrho = (np.sum(epsrho[self.layerMtrls] * self.layerWidths)
                           / sum(self.layerWidths))
        # IFR
        if not self.customIFR:
            self.ifrDelta, self.ifrLambda = self._get_IFRList()

    def _reset_for_basis(self, start: int, end: int):
        super()._reset_for_basis(start, end)
        self.layerMtrls = (self.layerMtrls*2)[start:end]
        self.layerDopings = (self.layerDopings*2)[start:end]

    def figure_of_merit(self, upper: int, lower: int) -> float:
        """Calculate Figure of Merit.
        This function must be called after solving for wave functions

        Parameters
        ----------
        upper, lower :
            define the transition from upper to lower

        Return
        -------
        float: Figure of Merit

        Yield
        ------
        tauLO_l : float
            the lower state lifetime from LO scattering
        tauLO_u : float
            the upper state lifetime from LO scattering
        tauLO_ul : float
            the transition rate from upper to lower due to LO scattering
        tauIFR_l : float
            the lower state lifetime from IFR scattering
        tauIFR_u : float
            the upper state lifetime from IFR scattering
        tauIFR_ul : float
            the transition rate from upper to lower due to IFR scattering
        tau_u : float
            1/(1/tauLO_u + 1/tauIFR_u)
        tau_l : float
            1/(1/tauLO_l + 1/tauIFR_l)
        FoM : float
            the Figure of Merit in unit angstrom^2 ps
        """
        self.tauLO_ul = 1/self.lo_transition(upper, lower)
        self.tauLO_l = self.lo_lifetime(lower)
        self.tauLO_u = self.lo_lifetime(upper)
        if self.includeIFR:
            self.tauIFR_ul = 1/self.ifr_transition(upper, lower)
            self.tauIFR_u = self.ifr_lifetime(upper)
            self.tauIFR_l = self.ifr_lifetime(lower)
            self.tau_u = 1/(1/self.tauLO_u + 1/self.tauIFR_u)
            self.tau_l = 1/(1/self.tauLO_l + 1/self.tauIFR_l)
            self.tau_ul = 1/(1/self.tauLO_ul + 1/self.tauIFR_ul)
        else:
            self.tau_u = self.tauLO_u
            self.tau_l = self.tauLO_l
            self.tau_ul = self.tauLO_ul
        return self.dipole(upper, lower)**2 * self.tau_u * (
            1 - self.tau_l / self.tau_ul)

    def effective_ridx(self, wl: ScalerOrArray) -> ScalerOrArray:
        """Return the effective refractive index for TM mode"""
        if sum(self.layerWidths) == 0:
            return 1.0
        self.mtrlRIdx = [(m.moleFrac * Material.rIdx[m.A.name](wl) +
                          (1 - m.moleFrac) * Material.rIdx[m.B.name](wl))
                         for m in self.mtrlAlloys]
        self.layerRIdx = np.array([self.mtrlRIdx[n] for n in self.layerMtrls])
        neff = np.average(1/self.layerRIdx**2, axis=0,
                          weights=self.layerWidths)**(-1/2)
        return neff

    @property
    def sheet_density(self) -> float:
        """Return the sheet density of doping per period, in unit cm^-2"""
        # 1E9 -> 1E17 cm^-3 * Angstrom -> cm^-2
        return sum(self.layerDopings[n] * self.layerWidths[n]
                   for n in range(0, len(self.layerWidths))) * 1E9

    def full_population(self) -> np.array:
        """Apart from SchrodingerLayer.full_population, this also yield
        current density, with knowledge of doping"""
        res = super().full_population()
        # ps^-1 -> kA/cm^2, where 1E9 = 1E12 * 1E-3, 1E12 is (ps^-1 -> s^-1)
        self.current = self.flow * self.sheet_density * e0 * 1E9
        return res

    def full_gain_spectrum(self, wl: ScalerOrArray = None) -> ScalerOrArray:
        """Perform fully automatic calculation for the gain on wavelength(s).
        """
        if wl is None:
            wl = self.wl
        neff = self.effective_ridx(wl)
        de0 = h * c0 / (wl * 1E-6) / e0
        gain = 0
        # self.gainul = {}
        for i in range(len(self.periodIdx)):
            upper = self.periodIdx[i]
            for j in range(i+1, len(self.periodIdx)):
                lower = self.periodIdx[j]
                for shift in (-1, 0, 1):
                    dipole = self._dipole(upper, lower, shift) * 1E-8  # to cm
                    Eu = self.eigenEs[upper]
                    El = self.eigenEs[lower] - shift * self.Eshift
                    de = Eu - El
                    dpop = self.population[i] - self.population[j]
                    if de < 0:
                        de, dpop = -de, -dpop
                    if self.includeIFR:
                        gamma_para = self._pGamma[shift][j][i]
                    else:
                        gamma_para = self.dephasing(upper, lower)
                        gamma_para /= (1E12 * hbar/e0)  # unit to Hz
                    gamma = (gamma_para +
                             (self.decayRates[i] + self.decayRates[j])/2
                             ) * 1E12 * hbar/e0
                    gainul = dpop * dipole**2 * gamma/(gamma**2 + (de-de0)**2)
                    # if np.max(np.abs(gainul)) > 4E-13:
                    #     self.gainul[(upper, lower)] = gainul
                    #     print(dpop * dipole**2, h*c0/(de*e0)*1E6, gamma/de,
                    #           upper, lower, shift)
                    gain = gain + gainul
        # e0^2 / (hbar * c0 * eps0) is dimension 1.. 1E-8 makes unit cm^-1
        gain *= e0**2*de0*self.sheet_density / (
                hbar*neff*c0*eps0*self.periodL*1E-8)
        return gain


def optimize_layer(qcl: QCLayers, n: int, upper: int, lower: int,
                   iter: int = 50):
    """Optimize FoM*Lorentzian for n-th layer thickness, assuming the state
    index does not change. optimization is performed by searching on the
    position resolution steps.

    Warning: This cannot specify a correct state if there are states
    index crossing.

    TODO: Use period recognizer to improve the algorithm
    """
    Eu = qcl.eigenEs[upper]
    El = qcl.eigenEs[lower]
    if Eu < El:
        upper, lower = lower, upper
        Eu, El = El, Eu
    # 1E-6 um -> m... in unit eV
    w0 = h * c0 / (qcl.wl * 1E-6) / e0

    def reduceFoM():
        wul = Eu - El
        gamma = qcl.dephasing(upper, lower)
        return qcl.figure_of_merit(upper, lower) * gamma * w0/(
            gamma**2 + (wul - w0)**2)
    width = round(qcl.layerWidths[n] / qcl.xres) * qcl.xres
    FoMnow = reduceFoM()
    print(("Start Optimizing Layer NO %d " % n) +
          ("for FoM between state %d and %d.\n" % (upper, lower)) +
          ("\tStart at width=%.1f, FoM=%.5g" % (width, FoMnow)))

    def newFoM(newWidth):
        qcl.layerWidths[n] = newWidth
        qcl.populate_x()
        qcl.solve_whole()
        qcl.dipole(upper, lower)
        return reduceFoM()
    FoMminus = newFoM(width - qcl.xres)
    FoMplus = newFoM(width + qcl.xres)
    for _ in range(iter):
        if FoMnow < FoMplus:
            FoMminus = FoMnow
            FoMnow = FoMplus
            width += qcl.xres
            FoMplus = newFoM(width + qcl.xres)
        elif FoMnow < FoMminus:
            FoMplus = FoMnow
            FoMnow = FoMminus
            width -= qcl.xres
            FoMminus = newFoM(width - qcl.xres)
        else:
            print("Maximum iteration reached.")
            break
        print("\twidth=%.1f, FoM=%.5g" % (width, FoMnow))
    qcl.layerWidths[n] = width
    print("finished, width=%.1f, FoM=%.5g" % (width, FoMnow))
    return FoMnow


def auto_gain(qcl: QCLayers, wls: ScalerOrArray = None):
    """Perform automatic gain calculation from newly loaded a qcl object.

    This is equivalent to

    .. code-block:: python

        qcl.populate_x()
        qcl.solve_whole()
        qcl.period_recognize()
        qcl.full_population()
        result = qcl.full_gain_spectrum(wls)
    """
    qcl.populate_x()
    qcl.solve_whole()
    qcl.period_recognize()
    qcl.full_population()
    return qcl.full_gain_spectrum(wls)


def optimize_global(qcl: QCLayers, iter: int = 50):
    """A global optimization using gradient descent."""
    layerNum = len(qcl.layerWidths)
    now = auto_gain(qcl)
    for n in range(iter):
        changed = False
        print("iteration {}:".format(n))
        print("    initial gain: {}".format(now))
        for n in range(layerNum):
            w = qcl.layerWidths[n]
            maxDiff = min(1, round(0.1*w/qcl.xres))
            qcl.layerWidths[n] = w - qcl.xres
            newMinus = auto_gain(qcl)
            qcl.layerWidths[n] = w + qcl.xres
            newPlus = auto_gain(qcl)
            print(("   layer No.{}.. Width w={}, "
                   "gain (0:{}, +:{}, -:{})").format(
                   n, w, now, newPlus, newMinus))
            if not (now > newMinus and now > newPlus):
                changed = True
                if now < newPlus and now < newMinus:
                    dw = 1 if newPlus > newMinus else -1
                else:
                    dif = (newPlus - newMinus)/2
                    ddif = newPlus + newMinus - 2*now
                    dw = -dif/ddif
                    if ddif > 0 or abs(dw) > maxDiff:
                        dw = maxDiff if dif > 0 else -maxDiff
                    elif round(dw) == 0:
                        dw = 1 if dw > 0 else -1
                    else:
                        dw = round(dw)
                qcl.layerWidths[n] = w + dw*qcl.xres
                print("    new width:", qcl.layerWidths[n])
                if dw == 1:
                    now = newPlus
                elif dw == -1:
                    now = newMinus
                else:
                    now = qcl._auto_gain()
        # TODO: optimize field
        if not changed:
            break
    print('Finished')



import typing

import numpy as np

floatOrArray = typing.Union[float, np.ndarray]


class PyBand:
    bandtype: str
    @classmethod
    def __init__(cls, *args, **kwargs) -> None: ...
    def __del__(self, *args, **kwargs) -> None: ...
    def __reduce__(self): ...


def cSimpleSolve1D(step: float, Es: np.ndarray, V: np.ndarray, m: floatOrArray,
                   xmin: int = 0, xmax: typing.Optional[int] = None
                   ) -> np.ndarray: ...


def cSimpleFillPsi(step: float, EigenEs: np.ndarray, V: np.ndarray,
                   m: floatOrArray, xmin: int = 0,
                   xmax: typing.Optional[int] = None) -> np.ndarray:
    """
    Find wave functions. Assume mass as given.
    """


def cBandSolve1D(step: float, Es: np.ndarray, V: np.ndarray, band: PyBand,
                 xmin: int = 0, xmax: typing.Optional[int] = None
                 ) -> np.ndarray:
    """
    Find eigen energies using band mass.
    """


def cBandFillPsi(step: float, EigenEs: np.ndarray, V: np.ndarray, band: PyBand,
                 xmin: int = 0, xmax: typing.Optional[int] = None,
                 Elower: typing.Optional[float] = None,
                 Eupper: typing.Optional[float] = None,
                 field: typing.Optional[float] = None) -> np.ndarray:
    """
    Find wave functions using band mass. `field`, `Elower` and `Eupper` is used
    only for bound the energy range of the wave functions: outside the bound
    the wavefunction is promised to be zero.
    """


def cLOphononScatter(step: float, kl: float,
                     psi_ij: np.ndarray, xmin: int = 0,
                     xmax: typing.Optional[int] = None) -> float: ...


def cLOtotal(step: float, kls: np.ndarray,
             psi_ijs: np.ndarray, fjs: np.ndarray): ...


def cisMP() -> bool: ...



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This file defines the Material class and Alloy class with material datas for
III/V semiconductor electron band. The module contains:

* **Material** class;
* **Alloy** class (inherited from Material class);
* **MParam** dictionary, which stores constants for different materials
* **AParam** dictionary, which stores constants for different alloys
"""

# ===========================================================================
# Reference
# [0]Vurgaftman, I., J. áR Meyer, and L. áR Ram-Mohan. "Band parameters for
#    III–V compound semiconductors and their alloys." Journal of applied
#    physics 89.11 (2001): 5815-5875
# [1]Handbook of Optics, Vol.2, ISBN: 0070479747
# [2]Herve P J L, Vandamme L K J. Empirical temperature dependence of the
#    refractive index of semiconductors[J]. Journal of Applied Physics,
#    1995, 77(10): 5476-5477.
# [3]Joachim Pipre, Semiconductor Optoelectronic Devices: Introduction to
#    Physics and Simulation, SIBN: 0080469787
# [4]Varshni, Y. P. (1967). Temperature dependence of the energy gap in
#    semiconductors. physica, 34(1), 149-154
# ===========================================================================

# from warnings import warn
# TODO: add warning for problematic datas
from numpy import sqrt

# This is to separate legacy fitting
from ErwinJr2.rFittings import AlGaAsIndex, SiNxIndex, SiO2Index

# flags for what effect to include
BOWING = True
VARSH = True

# for In0.53Ga0.47As, EcG = 0.22004154
#    use this as a zero point baseline. This is only here for consistency
#    with legacy code
bandBaseln = 0.22004154


class Material(object):
    """A semiconductor material class that stores material parameters

    Parameters
    ----------
    Name :
        Name of material
    Temperature :
        Temperature of the material
    """
    def __init__(self, Name: str, Temperature: float = 300):
        """

        Yields
        ------
        type : str
            Type of the material
        """
        self.name = Name
        self.param = MParam[self.name].copy()
        self.type = self.param.pop("Crystal")
        self.set_temperature(Temperature)

    def set_temperature(self, Temperature: float):
        """
        Set temperature of the material and update related parameters:
        lattice constant and band gap.

        Yields
        ------
        T : float
            Updated temperature
        param : dict
            lattice constant and band gap in this dictionary are updated
        """
        self.T = Temperature
        for k in self.param:
            if k.endswith("lc") and k+"_T" in self.param:
                # k is about lattice constant
                self.param[k] = (MParam[self.name][k] + (self.T - 300)
                                 * MParam[self.name][k+"_T"])
        # Varshni correction to bandgap [4]
        # major assumption:
        #   Correction only to conduction band (no influence on VBO)
        #   Although in reality the Varshni correction should be part
        #   conduction band, part valence
        if VARSH:
            for pt in ('G', 'X', 'L'):
                Varsh = -self.param["al"+pt] * self.T**2 / (
                    self.T + self.param["be"+pt])
                self.param['Eg'+pt] = MParam[self.name]['Eg'+pt] + Varsh

    def set_strain(self, a_parallel: float):
        """
        Update parameters' dependence on strain, according to Pikus-Bir
        interaction.

        Parameters
        ----------
        a_parallel :
            lattice constant of the substrate

        Yields
        ------
        eps_parallel : float
            Strain tensor within/parallel to the layer plane
        a_perp : float
            Lattice const. perpendicular to the layer plane
        eps_perp : float
            Strain tensor perpendicular to the layer plane
        param : dict
            Update parameters' dependence on strain
        """
        # TODO: crystal with more than one lattice constant?
        # eps_parallel: strain tensor within/parallel to the layer plane
        self.eps_parallel = a_parallel / self.param["alc"] - 1
        # self.a_perp: lattice const. perpendicular to the layer plane
        # Ask if MBE growers care about strain on monolayer thickness
        self.a_perp = self.param["alc"] * (
            1 - 2 * self.param["c12"] / self.param["c11"] * self.eps_parallel)
        # eps_perp: strain tensor perpendicular to the layer plane
        self.eps_perp = self.a_perp / self.param["alc"] - 1
        if self.type == "ZincBlende":
            Pec = (2 * self.eps_parallel + self.eps_perp) * self.param["acG"]
            Pev = (2 * self.eps_parallel + self.eps_perp) * self.param["av"]
            Qe = (- self.param["b"] * (self.param["c11"] + 2*self.param["c12"])
                  / self.param["c11"] * self.eps_parallel)
            self.param["EcG"] = (self.param["VBO"] + self.param["EgG"] + Pec
                                 - bandBaseln)
            self.param["EcL"] = (self.param["VBO"] + self.param["EgL"]
                                 + (2 * self.eps_parallel + self.eps_perp)
                                 * (self.param["acL"] + self.param["av"])
                                 - bandBaseln)
            self.param["EcX"] = (self.param["VBO"] + self.param["EgX"]
                                 + (2 * self.eps_parallel + self.eps_perp)
                                 * (self.param["acX"] + self.param["av"])
                                 + 2/3*self.param["XiX"]*(
                                    self.eps_perp - self.eps_parallel)
                                 - bandBaseln)
            self.param["ESO"] = sqrt(9 * Qe**2 + 2 * Qe * self.param["DSO"] +
                                     self.param["DSO"]**2)
            self.param["EgLH"] = (self.param["EgG"] + Pec + Pev - 1/2 * (
                Qe - self.param["DSO"] + self.param["ESO"]))
            self.param["EgSO"] = (self.param["EgG"] + Pec + Pev - 1/2 * (
                Qe - self.param["DSO"] - self.param["ESO"]))
            self.param["EvLH"] = self.param["EcG"] - self.param["EgLH"]
            self.param["EvSO"] = self.param["EcG"] - self.param["EgSO"]
        else:
            print("Warning: the strain effect on {} is not implemented".format(
                self.type
            ))
            self.param["EcG"] = (self.param["VBO"] + self.param["EgG"]
                                 - bandBaseln)


class Alloy(Material):
    """
    An alloy of material with Mole fraction x

    Parameters
    ----------
    Name :
        Name of the alloy
    x :
        The first composition's Mole fraction defined in AParam
    Temperature :
        Temperature of the alloy
    """
    def __init__(self, Name: str, x: float, Temperature: float = 300):
        self.name = Name
        self.comp = AParam[self.name]["composition"]
        self.A = Material(self.comp[0], Temperature)
        self.B = Material(self.comp[1], Temperature)
        assert self.A.type == self.B.type
        self.type = self.A.type
        # so alloy is A_x B_(1-x)
        self.moleFrac = x
        self.set_temperature(Temperature)

    def set_temperature(self, Temperature: float):
        """
        Set temperature of the alloy and update related parameters,
        lattice consant and band gap, by updating the temperature of the
        materials in the alloy


        Yields
        ------
        T : int
            Updated temperature
        """
        self.T = Temperature
        self.A.set_temperature(self.T)
        self.B.set_temperature(self.T)
        self.set_molefrac(self.moleFrac)

    def set_molefrac(self, x: float):
        """
        Update parameters of the alloy with Mole fraction x

        Parameters
        ----------
        x :
            The first composition's Mole fraction defined in AParam

        Yields
        ------
        param : dict
            stores the parameter of the alloy
        """
        self.moleFrac = x
        self.param = {}

        # For the Gamma band gap bowing in AlxGa1-xAs and AlxGa1-xSb
        # a linear or even a quadratic interpolation is not sufficient.
        # Here, the bowing parameter is not constant,
        # it depends on the alloy composition x.
        if self.name == 'AlGaAs':
            AParam[self.name]['EgG'] = -0.127 + 1.310 * x
        if self.name == 'AlGaSb':
            AParam[self.name]['EgG'] = -0.044 + 1.22 * x

        for k in self.A.param:
            #  if not k in B.param:
            #      continue
            self.param[k] = (x * self.A.param[k] + (1-x) * self.B.param[k])
            if BOWING and k in AParam[self.name]:
                # Bowing parameters
                self.param[k] -= x * (1-x) * AParam[self.name][k]


MParam = {
    # TODO: include Luttinger gamma
    "GaAs": {  # from Vurgaftman[0] unless specified
        "Crystal": "ZincBlende",
        # Lattice constant and thermal expansion
        "alc": 5.65325,       # Angstrom, Ref[3] Table 2.6
        "alc_T": 3.88e-5,     # Angstrom / K
        # Elastic stiffness constants
        'c11': 1221,   # GPa
        'c12': 566,    # GPa
        'EgG': 1.519,  # eV, band gap at Gamma at 0 K
        'EgL': 1.815,  # eV, band gap at L
        'EgX': 1.981,  # eV, band gap at X
        'VBO': -0.80,  # eV, valence band offset
        'DSO': 0.341,  # eV, Delta_SO from top of valence band
        # Pikus-Bir interaction parameters: strain correction
        'acG': -7.17,  # eV
        'acL': -4.91,  # eV, NextNano DB
        'acX': -0.16,  # eV, NextNano DB
        'av': -1.16,   # eV, valence band deformation potential
        'b': -2.0,     # eV
        'XiG': 0,      # eV
        'XiL': 6.5,    # eV, NextNano DB
        'XiX': 14.26,  # eV, NextNano DB
        'me0': 0.067,  # m0, effective mass at Gamma
        # Kane model parameters for effective mass
        'Ep': 28.8,    # eV
        'F': -1.94,    # unit 1
        # Varshni correction
        'alG': 0.5405e-3,   # eV/K, Varshni alpha (at Gamma point)
        'beG': 204,         # K, Varshni beta (at Gamma point)
        'alX': 0.460e-3,    # K, Varshni alpha at X
        'beX': 204,         # K, Varshni beta at X
        'alL': 0.605e-3,    # K, Varshni alpha at L
        'beL': 204,         # K, Varshni beta at L
        'epss': 12.9,       # static permittivity
        'epsInf': 10.86,    # high-frequency permittivity
        'hwLO': 35.3e-3     # LO phonon energy eV
    },

    'InAs': {  # from Vurgaftman[0] unless specified
        "Crystal": "ZincBlende",
        "alc": 6.0583, "alc_T": 2.74e-5,
        'c11': 832.9, 'c12': 452.6,
        'EgG': 0.417, 'EgL': 1.133, 'EgX': 1.433, 'VBO': -0.59, 'DSO': 0.39,
        'acG': -5.08, 'acL': -3.89,  # eV, NextNano DB
        'acX': -0.08,                # eV, NextNano DB
        'av': -1.00, 'b': -1.8, 'XiG': 0, 'XiL': 11.35, 'XiX': 3.7,
        'me0': 0.026, 'Ep': 21.5, 'F': -2.9,
        'alG': 0.276e-3, 'beG': 93,
        'alX': 0.276e-3, 'beX': 93,
        'alL': 0.276e-3, 'beL': 93,
        'epss': 14.3, 'epsInf': 11.6,
        'hwLO': 29.93e-3
    },

    'AlAs': {  # from Vurgaftman[0] unless specified
        "Crystal": "ZincBlende",
        "alc": 5.6611, "alc_T": 2.90e-5,
        'c11': 1250, 'c12': 534,
        'EgG': 3.099, 'EgL': 2.46, 'EgX': 2.24, 'VBO': -1.33, 'DSO': 0.28,
        'acG': -5.64, 'acL': -3.07,  # NextNano DB
        'acX': 2.54,                 # NextNano DB
        'av': -2.47, 'b': -2.3, 'XiG': 0, 'XiL': 11.35, 'XiX': 6.11,
        'me0': 0.15, 'Ep': 21.1, 'F': -0.48,
        'alG': 0.855e-3, 'beG': 530,
        'alX': 0.70e-3, 'beX': 530,
        'alL': 0.605e-3, 'beL': 204,
        'epss': 10.06, 'epsInf': 8.16,
        'hwLO': 49.8e-3
    },

    'AlSb': {  # from Vurgaftman[0] unless specified
        "Crystal": "ZincBlende",
        "alc": 6.1355, "alc_T": 2.60e-5,
        'c11': 876.9, 'c12': 434.1,
        'EgG': 2.386, 'EgL': 2.329, 'EgX': 1.696, 'VBO': -0.41, 'DSO': 0.676,
        'acG': -4.5, 'acL': 0,     # NextNano DB
        'acX': 2.54,               # NextNano DB
        'av': -1.4, 'b': -1.35, 'XiG': 0, 'XiL': 11.35, 'XiX': 6.11,
        'me0': 0.14, 'Ep': 18.7, 'F': -0.56,
        'alG': 0.42e-3, 'beG': 140,
        'alX': 0.39e-3, 'beX': 140,
        'alL': 0.58e-3, 'beL': 140,
        'epss': 12.04,    # ISBN 0849389127
        'epsInf': 10.24,  # ISBN 0849389127
        'hwLO': 42.7      # http://prb.aps.org/pdf/PRB/v43/i9/p7231_1
    },

    'GaSb': {  # from Vurgaftman[0] unless specified
        "Crystal": "ZincBlende",
        "alc": 6.0959, "alc_T": 4.72e-5,
        'c11': 884.2, 'c12': 402.6,
        'EgG': 0.812, 'EgL': 0.875, 'EgX': 1.141, 'VBO': -0.03, 'DSO': 0.76,
        'acG': -7.5,
        'av': -0.8, 'b': -2.0,
        'me0': 0.039, 'Ep': 27.0, 'F': -1.63,
        'alG': 0.417e-3, 'beG': 140,
        'alX': 0.475e-3, 'beX': 94,
        'alL': 0.597e-3, 'beL': 140,
    },

    'InSb': {  # from Vurgaftman[0] unless specified
        "Crystal": "ZincBlende",
        "alc": 6.4794, "alc_T": 3.48e-5,
        'c11': 684.7, 'c12': 373.5,
        'EgG': 0.235, 'EgL': 0.93, 'EgX': 0.63, 'VBO': 0, 'DSO': 0.81,
        'acG': -6.94,
        'av': -0.36, 'b': -2.0,
        'me0': 0.0135, 'Ep': 23.3, 'F': -0.23,
        'alG': 0.32e-3, 'beG': 170,
        # alX, beX, alL, beL are unknown... using gamma point data instead
        'alX': 0.32e-3, 'beX': 170,
        'alL': 0.32e-3, 'beL': 170,
    },

    'InP': {
        "Crystal": "ZincBlende",
        "alc": 5.869, "alc_T": 2.79e-5,
        'c11': 1011, 'c12': 561,
        'EgG': 1.4236, 'EgL': 2.014, 'EgX': 2.384,  # -3.7e-4 * T
        'VBO': -0.94, 'DSO': 0.108,
        'acG': -6.0, 'av': -0.6, 'b': -2.0,
        'alG': 0.363, 'beG': 162,
        'alL': 0.363, 'beL': 162,
        # alX, beX are unknown... using gamma point data instead
        'alX': 0.363, 'beX': 162,
        'me0': 0.0795, 'Ep': 20.7, 'F': -1.31,
    },

    "alpha-GaN": {
        "Crystal": "Wurtzite"
    },
    "beta-GaN": {
        "Crystal": "ZincBlende",
        # alc_T not given, use 0
        "alc": 4.50, "alc_T": 0,
        'c11': 293, 'c12': 159,
        'EgG': 3.299, 'EgL': 5.59, 'EgX': 4.52,
        'VBO': -2.64, 'DSO': 0.017,
        'acG': -2.2, 'av': -5.2, 'b': -2.2,
        'alG': 0.593e-3, 'beG': 600,
        'alL': 0.593e-3, 'beL': 600,
        'alX': 0.593e-3, 'beX': 600,
        'me0': 0.15, 'Ep': 25.0, 'F': -0.92,
    },

    "alpha-AlN": {
        "Crystal": "Wurtzite"
    },
    "beta-AlN": {
        "Crystal": "ZincBlende",
        # alc_T not given, use 0
        "alc": 4.38, "alc_T": 0,
        'c11': 304, 'c12': 160,
        'EgG': 4.9, 'EgL': 9.3, 'EgX': 6.0,
        'VBO': -3.44, 'DSO': 0.019,
        'acG': -6.0, 'av': -3.4, 'b': -1.9,
        'alG': 0.593e-3, 'beG': 600,
        'alL': 0.593e-3, 'beL': 600,
        'alX': 0.593e-3, 'beX': 600,
        'me0': 0.25, 'Ep': 27.1, 'F': 0.76,
    },

    "alpha-InN": {
        "Crystal": "Wurtzite"
    },
    "beta-InN": {
        "Crystal": "ZincBlende",
        # alc_T not given, use 0
        "alc": 4.98, "alc_T": 0,
        'c11': 187, 'c12': 125,
        'EgG': 1.94, 'EgL': 5.82, 'EgX': 2.51,
        'VBO': -2.38, 'DSO': 0.006,
        'acG': -1.85, 'av': -1.5, 'b': -1.2,
        'alG': 0.245e-3, 'beG': 624,
        'alL': 0.245e-3, 'beL': 624,
        'alX': 0.245e-3, 'beX': 624,
        'me0': 0.12, 'Ep': 25.0, 'F': -0.92,
    }


}

AParam = {
    'InGaAs': {
        'EgG': 0.477, 'EgL': 0.33, 'EgX': 1.4, 'VBO': -0.38, 'DSO': 0.15,
        'acG': 2.61, 'acL': 2.61,  # NextNano DB
        'acX': 2.61,               # NextNano DB
        'me0': 0.0091, 'Ep': -1.48, 'F': 1.77,
        'name': 'InxGa1-xAs',
        'composition': ('InAs', 'GaAs')
    },

    'AlInAs': {
        'EgG': 0.70, 'EgL': 0, 'EgX': 0, 'VBO': -0.64, 'DSO': 0.15,
        'acG': -1.4, 'acL': -1.4,  # NextNano DB
        'acX': -1.4,  # NextNano DB
        'me0': 0.049, 'Ep': -4.81, 'F': -4.44,
        'name': 'Al1-xInxAs',
        'composition': ('InAs', 'AlAs')
    },

    'AlGaAs': {
        'EgG': -0.127,  # + 1.310*x(Al)
        # To describe the band gap bowing at the Gamma point in AlxGa1-xAs,
        # a linear or even a quadratic interpolation is not sufficient.
        # Here, the bowing parameter is not constant,
        # it depends on the alloy composition x.
        'EgL': 0.055, 'EgX': 0, 'VBO': 0, 'DSO': 0,
        'acG': 0, 'acL': 0, 'acX': 0,
        'me0': 0, 'Ep': 0, 'F': 0,
        'name': 'AlxGa1-xAs',
        'composition': ('AlAs', 'GaAs')
    },

    'AlAsSb': {
        'EgG': 0.8, 'EgL': 0.28, 'EgX': 0.28, 'DSO': 0.15, 'VBO': -1.71,
        'name': 'AlAsxSb1-x',
        'composition': ('AlAs', 'AlSb')
    },

    'AlGaSb': {
        'EgG': -0.044,  # + 1.22*x(Al), same as AlGaAs
        'EgL': 0, 'EgX': 0, 'VBO': 0, 'DSO': 0.2,
        'acG': 0, 'acL': 0,  # NextNano DB
        'acX': 0,  # NextNano DB
        'me0': 0, 'Ep': 0, 'F': 0,
        'name': 'AlxGa1-xSb',
        'composition': ('AlSb', 'GaSb')
    },

    'InAsSb': {
        'EgG': 0.67, 'EgL': 0.6, 'EgX': 0.6, 'VBO': 0, 'DSO': 1.2,
        'acG': 0, 'acL': 0,  # NextNano DB
        'acX': 0,  # NextNano DB
        'me0': 0.035, 'Ep': 0, 'F': 0,
        'name': 'InAsxSb1-x',
        'composition': ('InAs', 'InSb')
    }
}


def main(material):
    print("Looking for parameters of %s:" % material)
    if material in AParam:
        A, B = AParam[material]['composition']
        print("Alloy with (%s)x(%s)1-x" % (A, B))
        print("%s: " % A, MParam[A])
        print("%s: " % B, MParam[B])
        print("Bowing parameters:", AParam[material])
        return 0
    elif material in MParam:
        print(MParam[material])
        return 0
    else:
        print("Not found.")
        return 1


# refractive indices
rIdx = {
    # J. Appl. Phys., 94, 6447-6455 (2003)
    # https://refractiveindex.info/?shelf=main&book=GaAs&page=Skauli
    # 0.97um - 17um
    "GaAs": lambda wl: sqrt(5.372514 + 5.466742/(1-(0.4431307/wl)**2)
                            + 0.02429960/(1-(0.8746453/wl)**2)
                            + 1.957522/(1-(36.9166/wl)**2)),
    # J. Appl. Phys. 36, 1841-1844 (1965)
    # https://refractiveindex.info/?shelf=main&book=InAs&page=Lorimor
    # 3.7um - 31.3um
    "InAs": lambda wl: sqrt(11.1 + 0.71/(1-(2.551/wl)**2)
                            + 2.75/(1-(45.66/wl)**2)),
    # Handbook of Optics, 2nd edition, Vol. 2. McGraw-Hill 1994
    # https://refractiveindex.info/?shelf=main&book=InP&page=Pettit
    # 0.95um - 10um
    "InP": lambda wl: sqrt(7.255 + 2.316/(1-(0.6263/wl)**2)
                           + 2.765/(1-(32.935/wl)**2)),
    # J. Appl. Phys. 42, 3499-3500 (1971)
    # https://refractiveindex.info/?shelf=main&book=AlAs&page=Fern
    # 0.56um - 2.2um
    "AlAs": lambda wl: sqrt(2.0792 + 6.0840/(1-(0.2822/wl)**2)
                            + 1.9/(1-(27.62/wl)**2)),
    "Au": lambda wl: (-0.1933-0.382j + (0.3321+6.8522j)*wl
                      + (0.0938-0.1289j)*wl**2),
    # from old ej
    # "Au": lambda wl: (-0.1933-0.382j + (0.3321+6.8522j)*wl
    #                   + (0.0938**2+0.1289**2*1j)*wl),
    "AlxGa1-xAs": lambda wl, x: AlGaAsIndex(wl, x),
    "SiNx": lambda wl: SiNxIndex(wl),
    "SiO2": lambda wl: SiO2Index(wl),
    "Air": lambda wl: 1
}


if __name__ == "__main__":
    import sys
    for material in sys.argv[1:]:
        main(material)
        print("")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import unittest

import numpy as np
from scipy.constants import e as e0
from scipy.constants import electron_mass as m0
from scipy.constants import hbar as hbar

from ErwinJr2.QCLayers import QCLayers, SchrodingerLayer

"""This unit test is to compare simulation with PhysRevB.50.8663"""


class GaInAs_AlInAs_Layer(SchrodingerLayer):
    """Sadly the parameters given in PhysRevB.50.8663 is not complete"""
    def __init__(self, xres, layerWidths, layerMtrl, EField=0):
        # fixed offset 0.51 eV
        Vc = {0: 0, 1: 0.51}
        layerVc = [Vc[m] for m in layerMtrl]
        self.layerMtrl = layerMtrl
        super().__init__(xres, layerWidths=layerWidths, EField=EField,
                         layerVc=layerVc)
        self.crystalType = 'ZincBlende'

    def layerMc(self, n):
        Mc = {0: 0.043, 1: 0.072}
        return Mc[self.layerMtrl[n]]

    def populate_material(self):
        N = self.xPoints.size
        self.xF = -0.5*np.ones(N)  # s.t. 1+2F = 0
        self.xEp = 18.3*np.ones(N)
        self.xESO = np.zeros(N)
        self.xEg = 0.79*np.ones(N)
        self.bandParams = (self.xEg, self.xF, self.xEp, self.xESO)


def validate_ODE(q):
    interface = (np.abs(np.diff(q.xVc)) > 0.1)
    interface = (interface[1:] | interface[:-1])
    residules = []
    for t in range(len(q.eigenEs)):
        E = q.eigenEs[t]
        psi = q.psis[t]
        xEg, xF, xEp, xESO = q.bandParams
        meff = 1/(1 + 2*xF + xEp/3*(1/(E-q.xVc+xEg+xESO) + 2/(E-q.xVc+xEg)))
        # interpolate half grid
        meff = (meff[1:] + meff[:-1])/2
        residule = (
            -np.diff(hbar**2/(2*m0*meff*e0)*np.diff(psi))/(q.xres*1E-10)**2
            + q.xVc[1:-1] * psi[1:-1] - E*psi[1:-1])
        residules.append(residule)
        interRes = residule[interface]
        # The interface has larger error
        np.testing.assert_almost_equal(interRes, np.zeros(interRes.shape),
                                       decimal=2)
        other = residule[~interface]
        np.testing.assert_almost_equal(other, np.zeros(other.shape))
    return np.array(residules)


class TestQuantumWell(unittest.TestCase):
    def test_single_well(self):
        # Test PhysRevB.50.8663 original model
        layers = [100, 52, 100]
        mtrls = [1, 0, 1]
        singleWell = GaInAs_AlInAs_Layer(0.01, layers, mtrls)
        singleWell.populate_x()
        singleWell.solve_whole()
        validate_ODE(singleWell)
        Ei = singleWell.eigenEs[1]
        Ej = singleWell.eigenEs[0]
        self.assertAlmostEqual((Ei-Ej) / (0.381-0.123), 1, 1)
        self.assertAlmostEqual(abs(singleWell.dipole(1, 0)) / 15.3, 1, 0)

        # Test QCLayers
        qcLayers = QCLayers(xres=0.01, layerWidths=layers,
                            layerMtrls=mtrls, repeats=1, T=10.0)
        qcLayers.populate_x()
        qcLayers.solve_whole()
        validate_ODE(qcLayers)
        Ei = qcLayers.eigenEs[1]
        Ej = qcLayers.eigenEs[0]
        self.assertAlmostEqual((Ei-Ej) / (0.381-0.123), 1, 1)
        self.assertAlmostEqual(abs(qcLayers.dipole(1, 0)) / 15.3, 1, 0)

    def test_double_well(self):
        layers = [200, 59, 13, 24, 200]
        mtrls = [1, 0, 1, 0, 1]
        qcLayers = QCLayers(xres=0.01, layerWidths=layers,
                            layerMtrls=mtrls, repeats=1, T=10.0)
        qcLayers.populate_x()
        qcLayers.solve_whole()
        validate_ODE(qcLayers)
        self.assertAlmostEqual(qcLayers.eigenEs[1] - qcLayers.eigenEs[0],
                               0.252 - 0.102, 2)
        self.assertAlmostEqual(qcLayers.eigenEs[2] - qcLayers.eigenEs[0],
                               0.373 - 0.102, 1)
        self.assertAlmostEqual(abs(qcLayers.dipole(1, 0)), 16.4, 0)

    def test_three_well(self):
        layers = [300, 46, 10, 20, 10, 19, 300]
        mtrls = [1, 0, 1, 0, 1, 0, 1]
        qcLayers = QCLayers(xres=0.01, layerWidths=layers,
                            layerMtrls=mtrls, repeats=1, T=10.0)
        qcLayers.populate_x()
        qcLayers.solve_whole()
        validate_ODE(qcLayers)
        self.assertAlmostEqual(qcLayers.eigenEs[1] - qcLayers.eigenEs[0],
                               0.242 - 0.126, 2)
        self.assertAlmostEqual(qcLayers.eigenEs[2] - qcLayers.eigenEs[0],
                               0.383 - 0.126, 2)
        self.assertAlmostEqual(qcLayers.eigenEs[3] - qcLayers.eigenEs[0],
                               0.494 - 0.126, 2)
        self.assertAlmostEqual(abs(qcLayers.dipole(1, 0)), 18.6, -1)


if __name__ == "__main__":
    unittest.main()



"""
This file defines class for plotting figures in ErwinJr
"""

import json
import os
import sys
from warnings import warn

import matplotlib
import numpy as np
from matplotlib.backend_bases import NavigationToolbar2, cursors
from matplotlib.backends.backend_qt5 import cursord  # , NavigationToolbar2QT
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
# import matplotlib.backends.qt_editor.figureoptions as figureoptions
from matplotlib.figure import Figure
from PyQt5.QtCore import QObject, Signal
from PyQt5.QtWidgets import QFileDialog, QMessageBox, QSizePolicy

matplotlib.use('Qt5Agg')
config = {
    "PlotMargin": {'l': 0.9, 'r': 0.12, 'b': 0.6, 't': 0.09},
    "fontsize": 12,
    "wfscale": 0.3,
    "modescale": 3,
    # This number not necessarily but is chosen to be consistent with the
    # default value of tol for QCLayers.periodRecognize as
    # wf_almost_zero / modescale = tol
    "wf_almost_zero": 1.5e-4,
}
if sys.platform == 'win32':
    config["PlotMargin"] = {'l': 0.9, 'r': 0.12, 'b': 0.55, 't': 0.09}
elif sys.platform == 'darwin':
    config["PlotMargin"] = {'l': 0.90, 'r': 0.12, 'b': 0.55, 't': 0.09}


def LoadConfig(fname="plotconfig.json"):
    try:
        with open(fname, 'r') as f:
            userConfig = json.load(f)
    except FileNotFoundError:
        userConfig = {}
        warn("Cannot load plot config file %s. Use default config." % fname,
             UserWarning)
    for k in userConfig:
        config[k] = userConfig[k]


class EJcanvas(FigureCanvas):
    """EJcanvas is designed for ErwinJr as a canvas for energy band and
    wavefunctions"""
    def __init__(self, xlabel='x', ylabel='y', parent=None):
        self.figure = Figure()
        super(EJcanvas, self).__init__(self.figure)
        #  NavigationToolbar2.__init__(self, self)
        self.setParent(parent)
        self.setMinimumWidth(200)
        self.setMinimumHeight(200)
        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.MinimumExpanding)
        self.updateGeometry()
        self.axes = self.figure.add_subplot(111)
        self.xlabel = xlabel
        self.ylabel = ylabel

    def set_axes(self, fsize=config["fontsize"]):
        self.axes.tick_params(axis='both', which='major', labelsize=fsize)
        self.axes.spines['top'].set_color('none')
        self.axes.spines['right'].set_color('none')
        self.axes.spines['bottom'].set_position(('outward', 5))
        self.axes.set_xlabel(self.xlabel, fontsize=fsize)
        self.axes.spines['left'].set_position(('outward', 5))
        self.axes.set_ylabel(self.ylabel, fontsize=fsize)
        self.axes.autoscale(enable=True, axis='x', tight=True)
        self.axes.autoscale(enable=True, axis='y', tight=True)

    def test(self):
        """A test function, plotting sin(x)"""
        x = np.linspace(0, 10, 100)
        self.axes.plot(x, np.sin(x))

    def resizeEvent(self, event):
        super(EJcanvas, self).resizeEvent(event)
        height = self.figure.get_figheight()
        width = self.figure.get_figwidth()
        margin = config["PlotMargin"]
        self.figure.subplots_adjust(
            left=margin['l'] / width, bottom=margin['b'] / height,
            right=1 - margin['r'] / width, top=1 - margin['t'] / height,
            wspace=0, hspace=0)

    def clear(self):
        self.axes.clear()
        self.set_axes()


class EJplotControl(NavigationToolbar2, QObject):
    """This class is an implementation of NavigationToolbar2 for controlling
    the canvas of ErwinJr plotting. The class is mainly inspired by
    `NavigationToolbar2QT` in backend_qt5.py
    Critical APIs are:
    - set_action to set necessary actions that are part of `toolitems` of the
      controller.
    - set_custom to set customized actions, basically 'layerselect' and
      `pairselect` for quantum wells/barriers and quantum states selection.
    - trigger_custom to trigger custom actions
    """
    message = Signal(str)

    def __init__(self, canvas, parent):
        # Not QToolBar to not draw any toolbar icons.
        QObject.__init__(self, parent)
        NavigationToolbar2.__init__(self, canvas)
        self._actions = {}
        self._custom_active = {}
        self._custom_cursor = {}
        self.zoomed = False

    def _init_toolbar(self):
        # override for backward compatibility.
        pass

    def set_action(self, callback, button):
        """According to matplotlib.backend_bases, supported actions are:
        toolitems = (
            ('Home', 'Reset original view', 'home', 'home'),
            ('Back', 'Back to  previous view', 'back', 'back'),
            ('Forward', 'Forward to next view', 'forward', 'forward'),
            (None, None, None, None),
            ('Pan',
            'Left button pans, Right button zooms\n'
            'x/y fixes axis, CTRL fixes aspect',
            'move', 'pan'),
            ('Zoom', 'Zoom to rectangle \nx/y fixes axis, CTRL fixes aspect',
            'zoom_to_rect', 'zoom'),
            ('Subplots', 'Configure subplots', 'subplots',
                'configure_subplots'),
            (None, None, None, None),
            ('Save', 'Save the figure', 'filesave', 'save_figure'),
          )
        format (text, tooltip_text, image_file, callback)
        """
        if callback in (self.toolitems[n][-1] for n in
                        range(len(self.toolitems))):
            button.clicked.connect(getattr(self, callback))
            self._actions[callback] = button
            if callback in ['zoom', 'pan']:
                button.setCheckable(True)
        else:
            raise TypeError("%s not supported." % callback)

    def set_custom(self, name, button, callback, cursor=cursors.HAND):
        """customized callback action,
        s.t. this class manages the button's check status
        and its interaction with canvas.
        Note that callback is called when clicked on canvas"""
        self._custom_active[name] = callback
        self._actions[name] = button
        self._custom_cursor[name] = cursor
        self._custom_mode = None
        self._custom_callBack = None
        button.setCheckable(True)

    def _get_mode_name(self):
        # This is a work around for backward compatibility due to
        # https://github.com/matplotlib/matplotlib/pull/17135
        if self._custom_mode:
            return self._custom_mode
        try:
            name = self.mode.name
        except AttributeError:
            name = self._active
        return name

    def _update_buttons_checked(self):
        # sync button check-states to match active mode
        if 'pan' in self._actions:
            self._actions['pan'].setChecked(self._get_mode_name() == 'PAN')
        if 'zoom' in self._actions:
            self._actions['zoom'].setChecked(self._get_mode_name() == 'ZOOM')
        for mode in self._custom_active:
            self._actions[mode].setChecked(self._get_mode_name() == mode)

    def _reset_mode(self):
        # This is a work around for self.mode value type being private.
        name = self._get_mode_name()
        if name == 'PAN':
            self.pan()
            return
        if name == 'ZOOM':
            self.zoom()
            return

    def pan(self, *args):
        # override
        super(EJplotControl, self).pan(*args)
        self._reset_custom()
        self._update_buttons_checked()

    def zoom(self, *args):
        # override
        super(EJplotControl, self).zoom(*args)
        self._reset_custom()
        self.zoomed = True
        self._update_buttons_checked()

    def home(self, *args):
        # override
        super(EJplotControl, self).home(*args)
        self.zoomed = False

    def _reset_custom(self):
        self._custom_mode = None
        if self._custom_callBack is not None:
            self.canvas.mpl_disconnect(self._custom_callBack)
            self._custom_callBack = None

    def trigger_custom(self, mode):
        self._reset_mode()
        if self._custom_mode != mode:
            self._reset_custom()
            self._custom_mode = mode
            self._custom_callBack = self.canvas.mpl_connect(
                'button_release_event', self._custom_active[mode])
            self.canvas.widgetlock(self)
        else:
            self._reset_custom()
            self.canvas.widgetlock.release(self)

        for a in self.canvas.figure.get_axes():
            a.set_navigate_mode(None)

        self._update_buttons_checked()

    def _update_cursor(self, event):
        # override backend_bases.NavigationToolbar2._update_cursor
        if (event.inaxes and self._custom_mode and
                self._lastCursor != self._custom_cursor[self._custom_mode]):
            self.set_cursor(self._custom_cursor[self._custom_mode])
            return
        super(EJplotControl, self)._update_cursor(event)

    def set_cursor(self, cursor):
        # implement
        self.canvas.setCursor(cursord[cursor])

    def draw_rubberband(self, event, x0, y0, x1, y1):
        # implement
        height = self.canvas.figure.bbox.height
        y1 = height - y1
        y0 = height - y0
        rect = [int(val) for val in (x0, y0, x1 - x0, y1 - y0)]
        self.canvas.drawRectangle(rect)

    def remove_rubberband(self):
        # implement
        self.canvas.drawRectangle(None)

    def save_figure(self, caption="Choose a filename to save to",
                    filename=None, default_filetype=None):
        # This method is override to support customized filename.
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = sorted(filetypes.items())
        if default_filetype is None:
            default_filetype = self.canvas.get_default_filetype()

        if not filename:
            startpath = os.path.expanduser(
                matplotlib.rcParams['savefig.directory'])
            filename = os.path.join(startpath,
                                    self.canvas.get_default_filename())
        filters = []
        selectedFilter = None
        for name, exts in sorted_filetypes:
            exts_list = " ".join(['*.%s' % ext for ext in exts])
            filter = '%s (%s)' % (name, exts_list)
            #  filter = '(*.%s) %s' % (exts_list, name)
            if default_filetype in exts:
                selectedFilter = filter
            filters.append(filter)
        filters = ';;'.join(filters)
        print(filters, selectedFilter)

        fname, filter = QFileDialog.getSaveFileName(
            self.parent(), caption, filename, filters, selectedFilter)
        if fname:
            try:
                self.canvas.figure.savefig(fname, dpi=300,
                                           bbox_inches='tight')
            except Exception as e:
                QMessageBox.critical(
                    self.parent(), "Error saving band diagram",
                    e, QMessageBox.Ok, QMessageBox.NoButton)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import sys
import traceback
from functools import partial

from PyQt5.QtCore import (QDir, QFile, QFileInfo, QSettings, QStandardPaths,
                          QUrl, QVariant)
from PyQt5.QtGui import QDesktopServices, QIcon, QKeySequence, QPixmap
from PyQt5.QtWidgets import (QAction, QApplication, QFileDialog, QInputDialog,
                             QMainWindow, QMessageBox, QSplashScreen,
                             QTabWidget)

from ErwinJr2 import SaveLoad
from ErwinJr2.OpticalTab import OpticalTab
from ErwinJr2.OptStrata import OptStrata
from ErwinJr2.QCLayers import QCLayers, onedq
from ErwinJr2.QuantumTab import QuantumTab
from ErwinJr2.versionAndName import VERSION

basePath = os.path.dirname(os.path.abspath(__file__))
DEFAULT_FILE_DIR = os.path.join(
    QStandardPaths.writableLocation(QStandardPaths.DocumentsLocation),
    'QCLDesign')


class MainWindow(QMainWindow):
    def __init__(self, fname=None, parent=None):
        super(MainWindow, self).__init__(parent)
        self.qsettings = QSettings(QSettings.IniFormat, QSettings.UserScope,
                                   "ErwinJr", "ErwinJr2", self)

        if self.qsettings.value('firstRun', True, type=bool):
            self.qsettings.setValue("firstRun", False)
            if not fname:
                firstRunBox = QMessageBox(
                    QMessageBox.Question, 'ErwinJr2 ' + VERSION,
                    ("Welcome to ErwinJr2!\n"
                     "Since this is your first time running the program, "
                     "would you like to open an example or a blank file?"),
                    parent=self)
                firstRunBox.addButton("Blank File", QMessageBox.NoRole)
                firstRunBox.addButton("Example File", QMessageBox.YesRole)
                ansr = firstRunBox.exec_()
                if ansr:
                    if not QDir(DEFAULT_FILE_DIR).exists():
                        os.makedirs(DEFAULT_FILE_DIR)
                    fname = os.path.join(DEFAULT_FILE_DIR, 'PQLiu.json')
                    if not QFile(fname).exists():
                        example = os.path.join(
                            basePath, 'example', 'PQLiu.json')
                        with open(example, 'r') as fin:
                            with open(fname, 'w') as fout:
                                fout.write(fin.read())
        elif not fname:
            # Load last file
            fname = self.qsettings.value("LastFile", None)
        self.dirty = True
        qclayers = None
        stratum = None
        rf = self.qsettings.value("RecentFiles")
        self.recentFiles = rf if rf else []
        if fname and QFile.exists(fname):
            try:
                with open(fname, 'r') as f:
                    qclayers, stratum = SaveLoad.loadBoth(f)
                self.filename = fname
                self.addRecentFile(fname)
                self.dirty = False
            except Exception:
                QMessageBox.warning(self, "ErwinJr2 - Warning",
                                    "Could not load the *.json file.\n" +
                                    traceback.format_exc())
                qclayers = None
                stratum = None
                self.filename = None
        else:
            self.filename = None

        self.mainTabWidget = QTabWidget()
        self._setTabs(qclayers, stratum)
        # self.mainTabWidget.setCurrentIndex(1)

        self.setCentralWidget(self.mainTabWidget)

        if self.qsettings.value("MainWindow/Geometry"):
            self.restoreGeometry(self.qsettings.value("MainWindow/Geometry"))
        if self.qsettings.value("MainWindow/State"):
            self.restoreState(self.qsettings.value("MainWindow/State"))
        self.updateFileMenu()
        self.updateWindowTitle()

    def _setTabs(self, qclayers, stratum):
        self.mainTabWidget.clear()
        # ==========================
        # Quantum Tab
        # ==========================
        self.qtab = QuantumTab(qclayers, self)
        self.qtab.dirty.connect(self.thingsChanged)
        self.mainTabWidget.addTab(self.qtab, 'Quantum')
        # ==========================
        # Optical Tab
        # ==========================
        self.otab = OpticalTab(stratum)
        self.otab.dirty.connect(self.thingsChanged)
        self.mainTabWidget.addTab(self.otab, 'Optics')
        self.qtab.toOpticsButton.clicked.connect(self.q2o)
        self.otab.fieldBox.setValue(self.qtab.qclayers.EField)
        self.create_menu()
        self.mainTabWidget.currentChanged.connect(self.create_menu)

    def q2o(self):
        self.otab.setupActive(self.qtab.wavelength, self.qtab.qclayers.EField,
                              self.qtab.gainCoeff, self.qtab.neff,
                              sum(self.qtab.qclayers.layerWidths))
        self.mainTabWidget.setCurrentIndex(1)

    def thingsChanged(self):
        """ SLOT connected to self.qtab.dirty"""
        self.dirty = True
        self.updateWindowTitle()

# ===========================================================================
# General Menu Functions
# ===========================================================================
    def add_actions(self, target, actions):
        for action in actions:
            if action is None:
                target.addSeparator()
            else:
                target.addAction(action)

    def create_action(self, text, slot=None, shortcut=None, icon=None,
                      tip=None, checkable=False, ischecked=False,
                      settingSync=False):
        action = QAction(text, self)
        if icon:
            action.setIcon(QIcon(
                os.path.join(basePath, 'images', '%s.png' % icon)))
        if shortcut:
            action.setShortcut(shortcut)
        if tip:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if slot:
            action.triggered.connect(slot)
        if checkable:
            action.setCheckable(True)
        if ischecked:
            action.setChecked(True)
        if settingSync:
            assert checkable
            action.triggered.connect(lambda: self.action_setting(text))
            # initialize
            if self.qsettings.value(text, ischecked, type=bool) != ischecked:
                action.setChecked(not ischecked)
                if slot:
                    slot()
        return action

    def action_setting(self, name):
        self.qsettings.setValue(
            name, not self.qsettings.value(name, False, type=bool))

    def create_menu(self, tabIndex=0):
        self.menuBar().clear()
        # file menu
        self.file_menu = self.menuBar().addMenu("&File")
        newFileAction = self.create_action(
            "&New...", slot=self.fileNew, shortcut=QKeySequence.New,
            icon="filenew", tip="New ErwinJr2 file")
        openFileAction = self.create_action(
            "&Open", shortcut="Ctrl+O", slot=self.fileOpen,
            tip="Open ErwinJr2 file", icon="fileopen")
        saveFileAction = self.create_action(
            "&Save", shortcut="Ctrl+S", slot=self.fileSave,
            tip="Save ErwinJr2 file", icon="filesave")
        saveAsFileAction = self.create_action(
            "S&ave As", shortcut="Ctrl+Alt+S", slot=self.fileSaveAs,
            tip="Save ErwinJr2 file as", icon="filesaveas")
        exportQuantumCanvasAction = self.create_action(
            "Export Band Diagram Image", slot=self.exportBandDiagram,
            tip="Export Band Diagram Image")
        exportBandCSVAction = self.create_action(
            "Export Band Diagram Data", slot=self.export_band_diagram_data,
            tip="Export Band Diagram Data")
        quit_action = self.create_action(
            "&Quit", slot=self.close, shortcut="Ctrl+W",
            tip="Close the application", icon="filequit")
        self.fileMenuActions = (
            newFileAction, openFileAction, saveFileAction, saveAsFileAction,
            None, exportBandCSVAction, exportQuantumCanvasAction, None,
            quit_action)
        self.add_actions(self.file_menu, self.fileMenuActions)
        self.file_menu.aboutToShow.connect(self.updateFileMenu)

        # edit menu
        if tabIndex == 0:
            self.edit_menu = self.menuBar().addMenu("&Edit")
            temperatureAction = self.create_action(
                "&Temperature", slot=self.set_temperature,
                tip="Set temperature")
            rotateLayerAction = self.create_action(
                "&Rotate Layer Table", slot=self.qtab.rotate_layer,
                tip="Move zeroth layer to first layer")
            rotateLayerAction.setShortcut("Ctrl+T")
            invertLayerAction = self.create_action(
                "&Invert Layer Table", slot=self.qtab.invert_layer,
                tip="Invert the layer order")
            solveARonly = self.create_action(
                "&Solve Active Only", checkable=True,
                ischecked=self.qtab.qclayers.basisAROnly,
                slot=self.qtab.ARonly)
            copyStructureAction = self.create_action(
                "&Copy Structure", slot=self.qtab.copy_structure,
                tip="Copy Layer Structure to Clipboard")
            self.add_actions(self.edit_menu, (temperatureAction,
                                              rotateLayerAction,
                                              invertLayerAction, None,
                                              solveARonly, None,
                                              copyStructureAction))

        # view menu
        self.view_menu = self.menuBar().addMenu("&View")
        if tabIndex == 0:
            VXBandAction = self.create_action(
                "X Valley Conduction Band", checkable=True,
                ischecked=self.qtab.plotVX,
                slot=self.qtab.view_VXBand)
            VLBandAction = self.create_action(
                "L Valley Conduction Band",
                checkable=True, ischecked=self.qtab.plotVL,
                slot=self.qtab.view_VLBand)
            LHBandAction = self.create_action(
                "Light Hole Valence Band",
                checkable=True, ischecked=self.qtab.plotLH,
                slot=self.qtab.view_LHBand)
            SOBandAction = self.create_action(
                "Split Off Valence Band",
                checkable=True, ischecked=self.qtab.plotSO,
                slot=self.qtab.view_SOBand)
            plotwf = self.create_action(
                "Plot Wave function",
                checkable=True, ischecked=self.qtab.plotType == 'wf',
                slot=self.qtab.set_plotwf, settingSync=True)
            plotFill = self.create_action(
                "Fill wave function curve",
                checkable=True, ischecked=bool(self.qtab.fillPlot),
                slot=self.qtab.set_fill, settingSync=True)
            self.add_actions(self.view_menu, (VXBandAction, VLBandAction,
                                              LHBandAction, SOBandAction,
                                              None, plotwf, plotFill,
                                              ))
        else:
            opticalActiceAction = self.create_action(
                "Optical: Show Active Core",
                checkable=True, ischecked=self.otab.redActive,
                slot=self.otab.view_redActive, settingSync=True)
            self.add_actions(self.view_menu, (opticalActiceAction, ))

        if tabIndex == 0:
            # model menu
            self.model_menu = self.menuBar().addMenu("&Model")
            self.solverActions = {}
            for solver in ("ODE", "matrix"):
                self.solverActions[solver] = self.create_action(
                    solver, checkable=True,
                    ischecked=self.qtab.qclayers.solver == solver,
                    slot=partial(self.choose_solver, solver)
                )
            if onedq is None:
                # C library does not exist
                self.solverActions['ODE'].setEnabled(False)
            solver_menu = self.model_menu.addMenu("Eigen Solver")
            solver_menu.addActions(self.solverActions.values())
            ifrAction = self.create_action(
                "IFR scattering", checkable=True,
                ischecked=self.qtab.qclayers.includeIFR,
                slot=self.qtab.triggerIFR)
            self.add_actions(self.model_menu, (None, ifrAction))

        # help menu
        # MacOS automatically remove the about item...
        aboutText = ("&Who write this?" if sys.platform.startswith('darwin')
                     else "&About")
        self.help_menu = self.menuBar().addMenu("&Help")
        about_action = self.create_action(aboutText, slot=self.on_about)
        licenses_action = self.create_action("&License",
                                             slot=self.on_licenses)
        tutorialAction = self.create_action("&Documents", shortcut='F1',
                                            slot=self.on_tutorial)
        self.add_actions(self.help_menu, (tutorialAction,
                                          about_action,
                                          licenses_action))

# ===========================================================================
# File Menu Items
# ===========================================================================
    def updateFileMenu(self):
        """SLOT connected to self.file_menu.aboutToShow()
        Update for recent files"""
        self.file_menu.clear()
        self.add_actions(self.file_menu, self.fileMenuActions[:-1])
        recentFiles = []
        for fname in self.recentFiles:
            if fname != self.filename and QFile.exists(fname):
                recentFiles.append(fname)
        if recentFiles:
            self.file_menu.addSeparator()
            for i, fname in enumerate(recentFiles):
                action = QAction(
                    "&{0}  {1}".format(i + 1,
                                       QFileInfo(fname).fileName()), self)
                action.setData(QVariant(fname))
                action.triggered.connect(partial(self.fileOpen, fname))
                self.file_menu.addAction(action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.fileMenuActions[-1])

# ===========================================================================
# Save, Load, RecentFiles
# ===========================================================================
    def updateWindowTitle(self):
        title = "ErwinJr2 " + VERSION
        if self.filename:
            title += " - %s" % os.path.basename(self.filename)
        if self.dirty:
            title += "[*]"
        self.setWindowTitle(title)
        self.setWindowModified(self.dirty)

    def addRecentFile(self, fname):
        if fname in self.recentFiles:
            self.recentFiles.pop(self.recentFiles.index(fname))
        self.recentFiles.insert(0, fname)
        while len(self.recentFiles) > 9:
            self.recentFiles.pop()

    def unsaveConfirm(self):
        """Confirm if unsaved data should be saved. This returns False if the
        user cancels the operation"""
        if self.dirty:
            reply = QMessageBox.question(
                self, "ErwinJr2 " + VERSION + " - Unsaved Changes",
                "Save unsaved changes?",
                QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
            if reply == QMessageBox.Cancel:
                return False
            elif reply == QMessageBox.Yes:
                self.fileSave()
        return True

    def fileNew(self):
        """Start a new file, confirm if there's unsaved data"""
        if not self.unsaveConfirm():
            return False
        self._setTabs(QCLayers(), OptStrata())
        self.filename = None
        self.dirty = True
        self.updateWindowTitle()
        return True

    def fileOpen(self, fname=None):
        """Clear all old data and load a new file. This will check
        self.unsaveConfirm and return False if user cancels it.
        This is used as user action and SLOT to fileOpen related signals."""
        if not self.unsaveConfirm():
            return False
        if not fname:
            fname, flt = QFileDialog.getOpenFileName(
                self, "ErwinJr2 - Choose file",
                os.path.dirname(self.filename) if self.filename else ".",
                "ErwinJr2 files (*.json)\nAll files (*.*)")
        if fname:
            self.qclLoad(fname)
            return True
        else:
            # User cancels the OpenFile dialog
            return False

    def qclLoad(self, fname):
        """Load from file "fname", and update everything for consistency."""
        try:
            with open(fname, 'r') as f:
                qclayers, stratum = SaveLoad.loadBoth(f)
                if stratum is None:
                    stratum = OptStrata(3.0)
        except Exception:
            QMessageBox.warning(self, "ErwinJr2 - Warning",
                                "Could not load the *.json file.\n" +
                                traceback.format_exc())
            return
        self._setTabs(qclayers, stratum)
        self.filename = fname
        self.addRecentFile(fname)
        self.dirty = False
        self.updateWindowTitle()

    def fileSave(self):
        if self.filename is None:
            return self.fileSaveAs()
        if self.filename.split('.')[-1] == 'json':
            try:
                with open(self.filename, 'w') as f:
                    SaveLoad.EJSaveJSON(f, self.qtab.qclayers,
                                        self.otab.stratum)
            except OSError:
                QMessageBox.warning(self, "ErwinJr2 - Warning",
                                    "Could not save *.json file.\n" +
                                    traceback.format_exc())
                return False
            self.dirty = False
            self.updateWindowTitle()
            return True
        else:
            raise IOError('The *.' + self.filename.split('.')[-1] +
                          ' extension is not supported.')

    def fileSaveAs(self):
        fname = self.filename if self.filename is not None else "."
        typeString = "ErwinJr2 file (*.json)\nAll files (*.*)"
        fname, flt = QFileDialog.getSaveFileName(
            self, "ErwinJr2 - Save File", fname, typeString)
        if fname:
            if "." not in fname:
                fname += ".json"
            self.addRecentFile(fname)
            self.filename = fname
            return self.fileSave()
        return False

    def closeEvent(self, event):
        if self.unsaveConfirm():
            filename = (QVariant(self.filename) if self.filename
                        else QVariant())
            self.qsettings.setValue("LastFile", filename)
            recentFiles = (QVariant(self.recentFiles) if self.recentFiles
                           else QVariant())
            self.qsettings.setValue("RecentFiles", recentFiles)
            self.qsettings.setValue(
                "MainWindow/Geometry", QVariant(self.saveGeometry()))
            self.qsettings.setValue(
                "MainWindow/State", QVariant(self.saveState()))
        else:
            event.ignore()

# ===========================================================================
# Export Functions
# ===========================================================================
    def exportBandDiagram(self):
        if self.filename is None:
            filename = 'new_qcl'
        else:
            filename = self.filename.split('.')[0]
        self.qtab.export_quantumCanvas(filename)

    def export_band_diagram_data(self):
        if self.filename is None:
            filename = 'new_qcl'
        else:
            filename = self.filename.split('.')[0]
        fname, flt = QFileDialog.getSaveFileName(
            self, "ErwinJr2 - Export Band Structure Data",
            filename,
            "Comma-Separated Value file (*.csv)")
        if fname:
            # if user doesn't click cancel
            self.qtab.export_band_data(fname)

# ===========================================================================
# Edit Menu Items
# ===========================================================================
    def set_temperature(self):
        nowTemp = self.qtab.qclayers.temperature
        newTemp, buttonResponse = QInputDialog.getDouble(
            self, 'ErwinJr2 Input Temperature', 'Set Temperature',
            value=nowTemp, min=0)
        if buttonResponse:
            self.qtab.set_temperature(newTemp)

# ===========================================================================
# Model Menu Items
# ===========================================================================
    def choose_solver(self, solver):
        if solver not in ("matrix", "ODE"):
            QMessageBox.warning(self, "Known solver: {}".format(solver)
                                + traceback.format_exc())
            return
        if solver != self.qtab.qclayers.solver:
            for action in self.solverActions.values():
                action.setChecked(False)
            self.solverActions[solver].setChecked(True)
            self.qtab.triggerSolver(solver)
            self.dirty = True
            self.thingsChanged()

# ===========================================================================
# Help Menu Items
# ===========================================================================
    def on_about(self):
        msg = """ErwinJr2 2.* and 1.0 Authors

        * Ming Lyu
            minglyu@princeton.edu

        With Contributions from:

        * Kevin Oresick (University of Wisconsin)

        ErwinJr2 0.x Authors and Contributors

         * Ming Lyu
         * Yaofeng (Desmond) Zhong
         * Xiaowen Chen

        ErwinJr 2.x Authors and Contributors

         * Kale J. Franz, PhD (Jet Propulsion Laboratory)
            kfranz@alumni.princeton.edu
            www.kalefranz.com

With contributions from:
         * Yamac Dikmelik (Johns Hopkins University)
         * Yu Song (Princeton University)
        """
        QMessageBox.about(self, "ErwinJr2 " + VERSION, msg.strip())

    def on_licenses(self):
        copyright = """
ErwinJr2 is a simulation program for quantum semiconductor lasers.
Copyright (C) 2021 Ming Lyu

The software inherits some design from ErwinJr, which is a work by Kale Franz.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
        QMessageBox.about(self, "ErwinJr2 " + VERSION,
                          copyright.strip())

    def on_tutorial(self):
        path = os.path.join(
            os.path.dirname(basePath), 'docs', '_build', 'html', 'index.html')
        if os.path.exists(path):
            QDesktopServices.openUrl(QUrl("file://" + os.path.abspath(path)))
        else:
            QDesktopServices.openUrl(
                QUrl("https://erwinjr2.readthedocs.io/en/stable/"))


def main(fileName=None):
    app = QApplication(sys.argv)
    app.setOrganizationName("ErwinJr")
    app.setOrganizationDomain("princetonuniversity.github.io/ErwinJr2")
    app.setApplicationName("ErwinJr2")
    app.setWindowIcon(QIcon(os.path.join(basePath, 'images', 'EJpng256.png')))

    # Create and display the splash screen
    splash_pix = QPixmap(os.path.join(basePath, 'images', 'splash.png'))
    splash = QSplashScreen(splash_pix)
    splash.setMask(splash_pix.mask())
    splash.show()

    form = MainWindow(fileName)
    form.show()
    splash.finish(form)
    if sys.platform.startswith('win'):
        # The default font for win is not consistent with the OS
        app.setFont(QApplication.font("QMenu"))
    app.exec_()


if __name__ == "__main__":
    # this block handles a filename passed in by command line
    try:
        fileName = os.path.abspath(sys.argv[1])
    except IndexError:
        fileName = None
    main(fileName)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import unittest

import numpy as np

from ErwinJr2.OptStrata import MaxwellLayer, MaxwellLayer_anisotropic


class TestTransferMatrix(unittest.TestCase):
    def test_real_bound_mode(self):
        """TM0 mode in [John Chilwell and Ian Hodgkinson,
        J. Opt. Soc. Am. A 1, 742-753 (1984)"""
        hs = np.array([500]*4)
        indices = np.array([1.0, 1.66, 1.53, 1.60, 1.66, 1.5])
        wl = 632.8
        stratum = MaxwellLayer(wl, hs, indices)
        beta = stratum.boundModeTM(max(indices))
        self.assertAlmostEqual(beta, 1.620031, 6)

        xs = np.linspace(-2000, 4000, 20000)
        n = stratum.populateIndices(xs)
        Ey, Hx, Ez = stratum.populateMode(beta, xs)
        # dHx/dy = i k n^2 Ez
        lhs = (Hx[2:] - Hx[:-2])/(xs[2]-xs[0])
        rhs = 1j*2*np.pi/wl*n**2*Ez
        np.testing.assert_almost_equal(lhs, rhs[1:-1], decimal=3)
        # n^2 d/dy n^-2 d/dy Hx = -(n^2k^2 - beta^2) Hx
        lhs = 1j*2*np.pi/wl*n[1:-1]**2*(Ez[2:] - Ez[:-2])/(xs[2]-xs[0])
        rhs = -(n**2 - beta**2)*(2*np.pi/wl)**2*Hx
        np.testing.assert_almost_equal(lhs, rhs[1:-1], decimal=3)
        self.assertAlmostEqual(Ez[0], 0, 5)
        self.assertAlmostEqual(Ez[-1], 0, 5)

    def test_anisotropy(self):
        hs = np.array([500]*4)
        indices = np.array([1.0, 1.66, 1.53, 1.60, 1.66, 1.5])
        wl = 632.8
        stratum = MaxwellLayer_anisotropic(wl, hs, indices)
        stratum.indexy[2] = 1.8
        beta = stratum.boundModeTM(max(indices))

        xs = np.linspace(-500, 2500, 10000)
        nz, ny = stratum.populateIndices(xs)
        np.testing.assert_raises(AssertionError,
                                 np.testing.assert_almost_equal, ny, nz)
        Ey, Hx, Ez = stratum.populateMode(beta, xs)
        # dHx/dy = i k nz^2 Ez
        lhs = (Hx[2:] - Hx[:-2])/(xs[2]-xs[0])
        rhs = 1j*2*np.pi/wl*nz**2*Ez
        np.testing.assert_almost_equal(lhs, rhs[1:-1], decimal=3)
        # ny^2 d/dy nz^-2 d/dy Hx = -(ny^2k^2 - beta^2) Hx
        lhs = 1j*2*np.pi/wl*ny[1:-1]**2*(Ez[2:] - Ez[:-2])/(xs[2]-xs[0])
        rhs = -(ny**2 - beta**2)*(2*np.pi/wl)**2*Hx
        np.testing.assert_almost_equal(lhs, rhs[1:-1], decimal=3)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
# -*- coding:utf-8 -*-
"""
This file is part of fftautocorr

Copyright (C) 2020 CareF
author: CareF
Licensed under a 3-clause BSD style license - see LICENSE.md

The script generates a table for the optimized padded length for FFT.
Running this script is part of pre-compilation. See Makefile for details.
"""

import os

import numpy as np


def tablegen(intmax, ps=(2, 3, 5)):
    """
    Generate a composite number table < intmax with prime composition in [ps]
    """
    logL = np.log2(intmax)
    allp = 1
    for p in ps:
        compp = p**(np.arange(logL/np.log2(p), dtype=np.int))
        allp = np.outer(allp, compp).reshape(-1)
        allp = allp[allp < intmax]
    return np.sort(allp)


def print_as_c_header(filename, array):
    contents = """/*
 * This file is generated by {this}
 */

#ifndef {filename}
#define {filename}
#include <stdlib.h>

static int find_factor(size_t n) {{
    const static int factortable[] = {{
        {numbers}
    }};
    int lo=0, hi=sizeof(factortable)/sizeof(factortable[0]), mid;
    while(lo < hi) {{
        mid = (lo + hi) / 2;
        if(factortable[mid] < n)
            lo = mid+1;
        else
            hi = mid;
        /* factortable[lo:] >= n; factortable[:lo] < n */
    }}
    if(lo == sizeof(factortable)/sizeof(factortable[0]))
        return -1;
    return factortable[lo];
}}
#endif
""".format(filename=filename.replace('.', '_').upper(),
           this=os.path.basename(__file__),
           numbers=', '.join([str(n) for n in array]))
    with open(filename, 'w') as f:
        f.write(contents)


if __name__ == "__main__":
    import sys
    filename = sys.argv[1]
    print("Generating %s" % filename)
    print_as_c_header(filename, tablegen(2**31-1))



from ctypes import CDLL, POINTER, Structure, c_double, c_int, c_void_p
from typing import Callable, Union

import numpy as np

__all__ = ['Band', 'cBandUpdateM', 'cBandNormalize']
from ErwinJr2.OneDQuantum.typeDefs import doubleArray


class cBand(Structure):
    _fields_ = [("updateM", c_void_p),
                ("normalize", c_void_p),
                ("N", c_int),
                ("Eg", POINTER(c_double))]


CBAND_P = POINTER(cBand)


class Band(object):
    """Python interface for a Band"""
    bandtype: str
    c: CBAND_P

    def __init__(self, bandtype: str, *args, **kwargs):
        super(Band, self).__init__()
        self.bandtype = bandtype
        # Make refs of parameters of band so it's not garbage collected
        self.args = args
        self.kwargs = kwargs
        if bandtype == "ZincBlende":
            self.c = cZBband_new(*args, **kwargs)
        else:
            raise ValueError(
                "crystal structure \"%s\" not implemented" % bandtype)

    def __del__(self):
        if self.bandtype == "ZincBlende":
            cZBband_free(self.c)
        else:
            raise ValueError(
                "crystal structure \"%s\" not implemented" % self.bandtype)


def cBandUpdateM(band: Band, E: float, V: np.ndarray, m: np.ndarray) -> int:
    raise NotImplementedError("Not linked to a proper clib.")


def cBandNormalize(band: Band, E: float, V: np.ndarray,
                   psi: np.ndarray, xres: float) -> float:
    raise NotImplementedError("Not linked to a proper clib.")


def cZBband_new(xEg: np.ndarray, xF: np.ndarray, xEp: np.ndarray,
                xESO: np.ndarray) -> CBAND_P:
    raise NotImplementedError("Not linked to a proper clib.")


def cZBband_free(zbband: CBAND_P) -> None:
    raise NotImplementedError("Not linked to a proper clib.")


def init(clib: CDLL) -> Union[
        Callable[[Band, float, np.ndarray, np.ndarray], int],
        Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], CBAND_P],
        Callable[[CBAND_P], None]]:
    global cBandUpdateM, cBandNormalize, cZBband_new, cZBband_free
    clib.BandUpdateM.argtypes = [POINTER(cBand), c_double,
                                 doubleArray, doubleArray]
    clib.BandUpdateM.restype = c_int
    clib.BandNormalize.argtypes = [POINTER(cBand), c_double, doubleArray,
                                   doubleArray, c_double]

    def cBandUpdateM(band: Band, E: float, V: np.ndarray, m: np.ndarray
                     ) -> int:
        return clib.BandUpdateM(band.c, c_double(E), V, m)

    def cBandNormalize(band: Band, E: float, V: np.ndarray,
                       psi: np.ndarray, xres: float) -> float:
        return clib.BandNormalize(band.c, c_double(E), V, psi, xres)

    clib.ZBband_new.argtypes = [c_int, doubleArray, doubleArray,
                                doubleArray, doubleArray]
    clib.ZBband_new.restype = POINTER(cBand)

    def cZBband_new(xEg: np.ndarray, xF: np.ndarray, xEp: np.ndarray,
                    xESO: np.ndarray) -> POINTER(cBand):
        return clib.ZBband_new(xEg.size, xEg, xF, xEp, xESO)

    clib.ZBband_free.argtypes = [POINTER(cBand)]
    clib.ZBband_free.restype = None

    def cZBband_free(zbband: POINTER(cBand)) -> None:
        return clib.ZBband_free(zbband)

    return cBandUpdateM, cZBband_new, cZBband_free
